# pylint:disable-msg=E0611,I1101
"""
Module bundling functions related to HTML and text processing,
content filtering and language detection.
"""

import gzip
import logging
import re
import zlib

from functools import lru_cache
from html import unescape
from itertools import islice
from unicodedata import normalize

# response compression
try:
    import brotli
    HAS_BROTLI = True
except ImportError:
    HAS_BROTLI = False

try:
    import zstandard
    HAS_ZSTD = True
except ImportError:
    HAS_ZSTD = False

# language detection
try:
    import py3langid
    LANGID_FLAG = True
except ImportError:
    LANGID_FLAG = False

# CChardet is faster and can be more accurate
try:
    from cchardet import detect as cchardet_detect
except ImportError:
    cchardet_detect = None
from charset_normalizer import from_bytes
from lxml.html import HtmlElement, HTMLParser, fromstring
# response types
from urllib3.response import HTTPResponse


LOGGER = logging.getLogger(__name__)

UNICODE_ALIASES = {'utf-8', 'utf_8'}

DOCTYPE_TAG = re.compile("^< ?! ?DOCTYPE.+?/ ?>", re.I)
FAULTY_HTML = re.compile(r"(<html.*?)\s*/>", re.I)

# note: htmldate could use HTML comments
# huge_tree=True, remove_blank_text=True
HTML_PARSER = HTMLParser(collect_ids=False, default_doctype=False, encoding='utf-8', remove_comments=True, remove_pis=True)

LINES_TRIMMING = re.compile(r'(?<![p{P}>])\n', flags=re.UNICODE|re.MULTILINE)

URL_BLACKLIST_REGEX = re.compile(r'^https?://|/+$')

# Regex to check image file extensions
IMAGE_EXTENSION = re.compile(r'[^\s]+\.(avif|bmp|gif|hei[cf]|jpe?g|png|webp)(\b|$)')

AUTHOR_PREFIX = re.compile(r'^([a-zäöüß]+(ed|t))? ?(written by|words by|words|by|von|from) ', flags=re.IGNORECASE)
AUTHOR_REMOVE_NUMBERS = re.compile(r'\d.+?$')
AUTHOR_TWITTER = re.compile(r'@[\w]+')
AUTHOR_REPLACE_JOIN = re.compile(r'[._+]')
AUTHOR_REMOVE_NICKNAME = re.compile(r'["‘({\[’\'][^"]+?[‘’"\')\]}]')
AUTHOR_REMOVE_SPECIAL = re.compile(r'[^\w]+$|[:()?*$#!%/<>{}~¿]')
AUTHOR_REMOVE_PREPOSITION = re.compile(r'\b\s+(am|on|for|at|in|to|from|of|via|with|—|-|–)\s+(.*)', flags=re.IGNORECASE)
AUTHOR_EMAIL = re.compile(r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b')
AUTHOR_SPLIT = re.compile(r'/|;|,|\||&|(?:^|\W)[u|a]nd(?:$|\W)', flags=re.IGNORECASE)
AUTHOR_EMOJI_REMOVE = re.compile(
    "["
    u"\U00002700-\U000027BF"  # Dingbats
    u"\U0001F600-\U0001F64F"  # Emoticons
    u"\U00002600-\U000026FF"  # Miscellaneous Symbols
    u"\U0001F300-\U0001F5FF"  # Miscellaneous Symbols And Pictographs
    u"\U0001F900-\U0001F9FF"  # Supplemental Symbols and Pictographs
    u"\U0001FA70-\U0001FAFF"  # Symbols and Pictographs Extended-A
    u"\U0001F680-\U0001F6FF"  # Transport and Map Symbols
    "]+", flags=re.UNICODE)
AUTHOR_REMOVE_HTML = re.compile(r'<[^>]+>')
CLEAN_META_TAGS = re.compile(r'["\']')

FORMATTING_PROTECTED = {'cell', 'head', 'hi', 'item', 'p', 'quote', 'ref', 'td'}
SPACING_PROTECTED = {'code', 'pre'}

# https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Language
TARGET_LANG_ATTRS = ('http-equiv="content-language"', 'property="og:locale"')
RE_HTML_LANG = re.compile(r'([a-z]{2})')

# Mostly filters for social media
RE_FILTER = re.compile(r'\W*(Drucken|E-?Mail|Facebook|Flipboard|Google|Instagram|'
                        'Linkedin|Mail|PDF|Pinterest|Pocket|Print|QQ|Reddit|Twitter|'
                        'WeChat|WeiBo|Whatsapp|Xing|Mehr zum Thema:?|More on this.{,8}$)$',
                       flags=re.IGNORECASE)
# COMMENTS_BLACKLIST = ('( Abmelden / Ändern )') # Fill in your details below|Trage deine Daten unten|Kommentar verfassen|Bitte logge dich|Hinterlasse einen Kommentar| to %s| mit %s)


def handle_compressed_file(filecontent):
    """
    Don't trust response headers and try to decompress a binary string
    with a cascade of installed packages. Use magic numbers when available.
    """
    if not isinstance(filecontent, bytes):
        return filecontent

    # source: https://stackoverflow.com/questions/3703276/how-to-tell-if-a-file-is-gzip-compressed
    if filecontent[:3] == b"\x1f\x8b\x08":
        try:
            return gzip.decompress(filecontent)
        except Exception:  # EOFError, OSError, gzip.BadGzipFile
            LOGGER.warning("invalid GZ file")
    # try zstandard
    if HAS_ZSTD and filecontent[:4] == b"\x28\xb5\x2f\xfd":
        try:
            return zstandard.decompress(filecontent)  # max_output_size=???
        except zstandard.ZstdError:
            LOGGER.warning("invalid ZSTD file")
    # try brotli
    if HAS_BROTLI:
        try:
            return brotli.decompress(filecontent)
        except brotli.error:
            pass  # logging.debug('invalid Brotli file')
    # try zlib/deflate
    try:
        return zlib.decompress(filecontent)
    except zlib.error:
        pass

    # return content unchanged if decompression failed
    return filecontent


def isutf8(data):
    """Simple heuristic to determine if a bytestring uses standard unicode encoding"""
    try:
        data.decode('UTF-8')
    except UnicodeDecodeError:
        return False
    return True


def detect_encoding(bytesobject):
    """"Read all input or first chunk and return a list of encodings"""
    # alternatives: https://github.com/scrapy/w3lib/blob/master/w3lib/encoding.py
    # unicode-test
    if isutf8(bytesobject):
        return ['utf-8']
    guesses = []
    # additional module
    if cchardet_detect is not None:
        cchardet_guess = cchardet_detect(bytesobject)['encoding']
        if cchardet_guess is not None:
            guesses.append(cchardet_guess.lower())
    # try charset_normalizer on first part, fallback on full document
    if len(bytesobject) < 10000:
        detection_results = from_bytes(bytesobject)
    else:
        detection_results = from_bytes(bytesobject[:5000] + bytesobject[-5000:]) or \
                            from_bytes(bytesobject)
    # return alternatives
    if len(detection_results) > 0:
        guesses.extend([r.encoding for r in detection_results])
    # it cannot be utf-8 (tested above)
    return [g for g in guesses if g not in UNICODE_ALIASES]


def decode_response(content):
    """Read the urllib3 object corresponding to the server response,
       try to guess its encoding and decode it to return a unicode string"""
    raise ValueError("decode_response() is deprecated, use decode_file() instead.")


def decode_file(filecontent):
    """Check if the bytestring could be GZip and eventually decompress it,
       guess bytestring encoding and try to decode to Unicode string.
       Resort to destructive conversion otherwise."""
    # init
    if isinstance(filecontent, str):
        return filecontent
    htmltext = None
    # GZip and Brotli test
    filecontent = handle_compressed_file(filecontent)
    # encoding
    for guessed_encoding in detect_encoding(filecontent):
        try:
            htmltext = filecontent.decode(guessed_encoding)
        except (LookupError, UnicodeDecodeError): # VISCII: lookup
            LOGGER.warning('wrong encoding detected: %s', guessed_encoding)
            htmltext = None
        else:
            break
    # return original content if nothing else succeeded
    return htmltext or str(filecontent, encoding='utf-8', errors='replace')


def is_dubious_html(beginning: str) -> bool:
    "Assess if the object is proper HTML (awith a corresponding tag or declaration)."
    return "html" not in beginning


def repair_faulty_html(htmlstring: str, beginning: str) -> str:
    "Repair faulty HTML strings to make then palatable for libxml2."
    # libxml2/LXML issue: https://bugs.launchpad.net/lxml/+bug/1955915
    if "doctype" in beginning:
        firstline, _, rest = htmlstring.partition("\n")
        htmlstring = DOCTYPE_TAG.sub("", firstline, count=1) + "\n" + rest
    # other issue with malformed documents: check first three lines
    for i, line in enumerate(iter(htmlstring.splitlines())):
        if "<html" in line and line.endswith("/>"):
            htmlstring = FAULTY_HTML.sub(r"\1>", htmlstring, count=1)
            break
        if i > 2:
            break
    return htmlstring


def fromstring_bytes(htmlobject):
    "Try to pass bytes to LXML parser."
    tree = None
    try:
        tree = fromstring(htmlobject.encode("utf8", "surrogatepass"), parser=HTML_PARSER)
    except Exception as err:
        LOGGER.error("lxml parser bytestring %s", err)
    return tree


def load_html(htmlobject):
    """Load object given as input and validate its type
    (accepted: lxml.html tree, trafilatura/urllib3 response, bytestring and string)
    """
    # use tree directly
    if isinstance(htmlobject, HtmlElement):
        return htmlobject
    # use trafilatura or urllib3 responses directly
    if isinstance(htmlobject, HTTPResponse) or hasattr(htmlobject, "data"):
        htmlobject = htmlobject.data
    # do not accept any other type after this point
    if not isinstance(htmlobject, (bytes, str)):
        raise TypeError("incompatible input type", type(htmlobject))
    # start processing
    tree = None
    # try to guess encoding and decode file: if None then keep original
    htmlobject = decode_file(htmlobject)
    # sanity checks
    beginning = htmlobject[:50].lower()
    check_flag = is_dubious_html(beginning)
    # repair first
    htmlobject = repair_faulty_html(htmlobject, beginning)
    # first pass: use Unicode string
    fallback_parse = False
    try:
        tree = fromstring(htmlobject, parser=HTML_PARSER)
    except ValueError:
        # "Unicode strings with encoding declaration are not supported."
        tree = fromstring_bytes(htmlobject)
        fallback_parse = True
    except Exception as err:  # pragma: no cover
        LOGGER.error("lxml parsing failed: %s", err)
    # second pass: try passing bytes to LXML
    if (tree is None or len(tree) < 1) and not fallback_parse:
        tree = fromstring_bytes(htmlobject)
    # rejection test: is it (well-formed) HTML at all?
    # log parsing errors
    if tree is not None and check_flag is True and len(tree) < 2:
        LOGGER.error(
            "parsed tree length: %s, wrong data type or not valid HTML", len(tree)
        )
        tree = None
    return tree


@lru_cache(maxsize=2**14)  # sys.maxunicode = 1114111
def return_printables_and_spaces(char):
    'Return a character if it belongs to certain classes'
    return char if char.isprintable() or char.isspace() else ''


def remove_control_characters(string):
    '''Prevent non-printable and XML invalid character errors'''
    return ''.join(map(return_printables_and_spaces, string))


def normalize_unicode(string, unicodeform='NFC'):
    'Normalize the given string to the specified unicode format.'
    return normalize(unicodeform, string)


@lru_cache(maxsize=1024)
def line_processing(line, preserve_space=False, trailing_space=False):
    '''Remove HTML space entities, then discard incompatible unicode
       and invalid XML characters on line level'''
    # spacing HTML entities: https://www.w3.org/MarkUp/html-spec/html-spec_13.html
    # unique code spaces
    new_line = remove_control_characters(line.replace('&#13;', '\r').replace('&#10;', '\n').replace('&nbsp;', '\u00A0'))
    if not preserve_space:
        # remove newlines that are not related to punctuation or markup
        # remove non-printable chars and normalize space characters (including Unicode spaces)
        new_line = trim(LINES_TRIMMING.sub(r" ", new_line))
        # prune empty lines
        if all(map(str.isspace, new_line)):
            new_line = None
        elif trailing_space:
            space_before = " " if line[0].isspace() else ""
            space_after = " " if line[-1].isspace() else ""
            new_line = "".join([space_before, new_line, space_after])
    return new_line


def sanitize(text, preserve_space=False, trailing_space=False):
    '''Convert text and discard incompatible and invalid characters'''
    # consider all text as a single line
    if trailing_space:
        return line_processing(text, preserve_space, True)
    # process line by line
    try:
        return '\n'.join(filter(None, (line_processing(l, preserve_space) for l in text.splitlines()))).replace('\u2424', '')
    except AttributeError:
        return None


def sanitize_tree(tree):
    '''Trims spaces, removes control characters and normalizes unicode'''
    for elem in tree.iter():
        parent = elem.getparent()
        parent_tag = parent.tag if parent is not None else ""

        # preserve space if the element or its parent is a specific tag, or if the element has text and children
        # the last part is relevant for item elements with ref inside for example
        preserve_space = elem.tag in SPACING_PROTECTED or parent_tag in SPACING_PROTECTED
        trailing_space = elem.tag in FORMATTING_PROTECTED or parent_tag in FORMATTING_PROTECTED or preserve_space

        # remove invalid attributes
        for attribute in elem.attrib:
            if ':' in attribute:  # colon is reserved for namespaces in XML
                if not elem.attrib[attribute] or attribute.split(':', 1)[0] not in tree.nsmap:
                    elem.attrib.pop(attribute)

        if elem.text:
            elem.text = sanitize(elem.text, preserve_space, trailing_space)
        if elem.tail:
            elem.tail = sanitize(elem.tail, preserve_space, trailing_space)
    return tree


@lru_cache(maxsize=1024)
def trim(string):
    '''Remove unnecessary spaces within a text string'''
    try:
        # remove newlines that are not related to punctuation or markup + proper trimming
        # return LINES_TRIMMING.sub(r' ', string).strip(' \t\n\r\v')
        # faster:
        return ' '.join(string.split()).strip()
    except (AttributeError, TypeError):
        return None


def normalize_tags(tags):
    '''Remove special characters of tags'''
    tags = CLEAN_META_TAGS.sub(r'', trim(unescape(tags)))
    return ", ".join(filter(None, tags.split(", ")))


def is_image_file(imagesrc):
    '''Check if the observed string corresponds to a valid image extension,
       return False otherwise'''
    return bool(imagesrc is not None and IMAGE_EXTENSION.search(imagesrc))


def normalize_authors(current_authors, author_string):
    '''Normalize author info to focus on author names only'''
    new_authors = []
    if author_string.lower().startswith('http') or AUTHOR_EMAIL.match(author_string):
        return current_authors
    if current_authors is not None:
        new_authors = current_authors.split('; ')
    # fix to code with unicode
    if '\\u' in author_string:
        author_string = author_string.encode().decode('unicode_escape')
    # fix html entities
    if '&#' in author_string or '&amp;' in author_string:
        author_string = unescape(author_string)
    # remove html tags
    author_string = AUTHOR_REMOVE_HTML.sub('', author_string)
    # examine names
    for author in AUTHOR_SPLIT.split(author_string):
        author = trim(author)
        # remove emoji
        author = AUTHOR_EMOJI_REMOVE.sub('', author)
        # remove @username
        author = AUTHOR_TWITTER.sub('', author)
        # replace special characters with space
        author = trim(AUTHOR_REPLACE_JOIN.sub(' ', author))
        author = AUTHOR_REMOVE_NICKNAME.sub('', author)
        # remove special characters
        author = AUTHOR_REMOVE_SPECIAL.sub('', author)
        author = AUTHOR_PREFIX.sub('', author)
        author = AUTHOR_REMOVE_NUMBERS.sub('', author)
        author = AUTHOR_REMOVE_PREPOSITION.sub('', author)
        # skip empty or improbably long strings
        if len(author) == 0 or (
            # simple heuristics, regex or vowel tests also possible
            ' ' not in author and '-' not in author and len(author) >= 50
            ):
            continue
        # title case
        if not author[0].isupper() or sum(1 for c in author if c.isupper()) < 1:
            author = author.title()
        # safety checks
        if author not in new_authors and (len(new_authors) == 0 or all(new_author not in author for new_author in new_authors)):
            new_authors.append(author)
    if len(new_authors) == 0:
        return current_authors
    return '; '.join(new_authors).strip('; ')


def make_chunks(iterable, n):
    """
    Chunk data into smaller pieces.
    https://docs.python.org/3/library/itertools.html
    """
    it = iter(iterable)
    while True:
        chunk = tuple(islice(it, n))
        if not chunk:
            return
        yield chunk
    # Python 3.8+ with walrus operator
    # while batch := tuple(islice(it, n)):
    #    yield batch


def is_acceptable_length(my_len, options):
    "Check if the document length is within acceptable boundaries."
    if my_len < options.min_file_size:
        LOGGER.error("too small/incorrect for URL %s", options.url)
        return False
    if my_len > options.max_file_size:
        LOGGER.error("too large: length %s for URL %s", my_len, options.url)
        return False
    return True


def check_html_lang(tree, target_language, strict=False):
    """Check HTML meta-elements for language information and split
       the result in case there are several languages."""
    for attr in TARGET_LANG_ATTRS:
        elems = tree.findall(f'.//meta[@{attr}][@content]')
        if elems:
            if any(target_language in RE_HTML_LANG.split(elem.get("content", "").lower()) for elem in elems):
                return True
            LOGGER.debug("%s lang attr failed", attr)
            return False

    # HTML lang attribute: sometimes a wrong indication
    if strict:
        elems = tree.xpath("//html[@lang]")
        if elems:
            if any(target_language in RE_HTML_LANG.split(elem.get("lang", "").lower()) for elem in elems):
                return True
            LOGGER.debug("HTML lang failed")
            return False

    LOGGER.debug("No relevant lang elements found")
    return True


def language_classifier(temp_text, temp_comments):
    '''Run external component (if installed) for language identification'''
    if LANGID_FLAG is True:
        result, _ = (
            py3langid.classify(temp_text)
            if len(temp_text) > len(temp_comments)
            else py3langid.classify(temp_comments)
        )
    else:
        LOGGER.warning('Language detector not installed, skipping detection')
        result = None
    return result


def language_filter(temp_text, temp_comments, target_language, docmeta):
    '''Filter text based on language detection and store relevant information'''
    # todo: run and pass info along anyway?
    if target_language is not None:
        # more thorough: detection on actual text content
        docmeta.language = language_classifier(temp_text, temp_comments)
        # HTML lang check? sometimes contradicted by detection above
        #if docmeta.language is None:
        #    if check_html_lang(tree, target_language) is False:
        #        LOGGER.error('wrong HTML meta language for URL %s', url)
        #        raise ValueError
        if docmeta.language is not None and docmeta.language != target_language:
            LOGGER.warning('wrong language: %s %s', docmeta.language, docmeta.url)
            return True, docmeta
    return False, docmeta


def textfilter(element):
    '''Filter out unwanted text'''
    testtext = element.tail if element.text is None else element.text
    # to check: line len → continue if len(line) <= 5
    return not text_chars_test(testtext) or any(map(RE_FILTER.match, testtext.splitlines()))


def text_chars_test(string):
    '''Determine if a string is only composed of spaces and/or control characters'''
    # or not re.search(r'\w', string)
    # return string is not None and len(string) != 0 and not string.isspace()
    return bool(string) and not string.isspace()



"""Handle the <a> tag."""
from typing import Dict

from haruka_parser.table_processing.html_state import HtmlDocumentState


def a_start_handler(state: HtmlDocumentState, attrs: Dict) -> None:
    """Handle the <a> tag."""
    state.link_target = ""
    if state.config.display_links:
        state.link_target = attrs.get("href", "")
    if state.config.display_anchors:
        state.link_target = state.link_target or attrs.get("name", "")

    if state.link_target:
        state.tags[-1].write("[")


def a_end_handler(state: HtmlDocumentState) -> None:
    """Handle the </a> tag."""
    if state.link_target:
        state.tags[-1].write(f"]({state.link_target})")



"""
Functions needed to scrape metadata from JSON-LD format.
For reference, here is the list of all JSON-LD types: https://schema.org/docs/full.html
"""

import json
import re
from html import unescape

from .utils import normalize_authors, trim

JSON_ARTICLE_SCHEMA = {"article", "backgroundnewsarticle", "blogposting", "medicalscholarlyarticle", "newsarticle", "opinionnewsarticle", "reportagenewsarticle", "scholarlyarticle", "socialmediaposting", "liveblogposting"}
JSON_OGTYPE_SCHEMA = {"aboutpage", "checkoutpage", "collectionpage", "contactpage", "faqpage", "itempage", "medicalwebpage", "profilepage", "qapage", "realestatelisting", "searchresultspage", "webpage", "website", "article", "advertisercontentarticle", "newsarticle", "analysisnewsarticle", "askpublicnewsarticle", "backgroundnewsarticle", "opinionnewsarticle", "reportagenewsarticle", "reviewnewsarticle", "report", "satiricalarticle", "scholarlyarticle", "medicalscholarlyarticle", "socialmediaposting", "blogposting", "liveblogposting", "discussionforumposting", "techarticle", "blog", "jobposting"}
JSON_PUBLISHER_SCHEMA = {"newsmediaorganization", "organization", "webpage", "website"}
JSON_AUTHOR_1 = re.compile(r'"author":[^}[]+?"name?\\?": ?\\?"([^"\\]+)|"author"[^}[]+?"names?".+?"([^"]+)', re.DOTALL)
JSON_AUTHOR_2 = re.compile(r'"[Pp]erson"[^}]+?"names?".+?"([^"]+)', re.DOTALL)
JSON_AUTHOR_REMOVE = re.compile(r',?(?:"\w+":?[:|,\[])?{?"@type":"(?:[Ii]mageObject|[Oo]rganization|[Ww]eb[Pp]age)",[^}[]+}[\]|}]?')
JSON_PUBLISHER = re.compile(r'"publisher":[^}]+?"name?\\?": ?\\?"([^"\\]+)', re.DOTALL)
JSON_TYPE = re.compile(r'"@type"\s*:\s*"([^"]*)"', re.DOTALL)
JSON_CATEGORY = re.compile(r'"articleSection": ?"([^"\\]+)', re.DOTALL)
JSON_NAME = re.compile(r'"@type":"[Aa]rticle", ?"name": ?"([^"\\]+)', re.DOTALL)
JSON_HEADLINE = re.compile(r'"headline": ?"([^"\\]+)', re.DOTALL)
JSON_MATCH = re.compile(r'"author":|"person":', flags=re.IGNORECASE)
JSON_REMOVE_HTML = re.compile(r'<[^>]+>')
JSON_SCHEMA_ORG = re.compile(r"^https?://schema\.org", flags=re.IGNORECASE)
JSON_UNICODE_REPLACE = re.compile(r'\\u([0-9a-fA-F]{4})')


def extract_json(schema, metadata):
    '''Parse and extract metadata from JSON-LD data'''
    if isinstance(schema, dict):
        schema = [schema]

    for parent in filter(lambda p: '@context' in p and isinstance(p['@context'], str) and JSON_SCHEMA_ORG.match(p['@context']), schema):
        if '@graph' in parent:
            parent = parent['@graph'] if isinstance(parent['@graph'], list) else [parent['@graph']]
        elif '@type' in parent and isinstance(parent['@type'], str) and 'liveblogposting' in parent['@type'].lower() and 'liveBlogUpdate' in parent:
            parent = parent['liveBlogUpdate'] if isinstance(parent['liveBlogUpdate'], list) else [parent['liveBlogUpdate']]
        else:
            parent = schema

        for content in filter(None, parent):
            # try to extract publisher
            if 'publisher' in content and 'name' in content['publisher']:
                metadata.sitename = content['publisher']['name']

            if '@type' not in content or len(content["@type"]) == 0:
                continue
            if isinstance(content["@type"], list):
                # some websites are using ['Person'] as type
                content_type = content["@type"][0].lower()
            else:
                content_type = content["@type"].lower()

            # The "pagetype" should only be returned if the page is some kind of an article, category, website...
            if content_type in JSON_OGTYPE_SCHEMA and not metadata.pagetype:
                metadata.pagetype = normalize_json(content_type)

            if content_type in JSON_PUBLISHER_SCHEMA:
                candidate = next((content[candidate] for candidate in ("name", "legalName", "alternateName") if content.get(candidate)), None)
                if candidate and isinstance(candidate, str):
                    if metadata.sitename is None or (len(metadata.sitename) < len(candidate) and content_type != "webpage"):
                        metadata.sitename = candidate
                    if metadata.sitename is not None and metadata.sitename.startswith('http') and not candidate.startswith('http'):
                        metadata.sitename = candidate

            elif content_type == "person":
                if content.get('name') and not content['name'].startswith('http'):
                    metadata.author = normalize_authors(metadata.author, content['name'])

            elif content_type in JSON_ARTICLE_SCHEMA:
                # author and person
                if 'author' in content:
                    list_authors = content['author']
                    if isinstance(list_authors, str):
                        # try to convert to json object
                        try:
                            list_authors = json.loads(list_authors)
                        except json.JSONDecodeError:
                            # it is a normal string
                            metadata.author = normalize_authors(metadata.author, list_authors)

                    if not isinstance(list_authors, list):
                        list_authors = [list_authors]
                    for author in list_authors:
                        if '@type' not in author or author['@type'] == 'Person':
                            # error thrown: author['name'] can be a list (?)
                            if 'name' in author and author['name'] is not None:
                                author_name = author['name']
                                if isinstance(author_name, list):
                                    author_name = '; '.join(author_name).strip('; ')
                                elif isinstance(author_name, dict) and "name" in author_name:
                                    author_name = author_name["name"]
                                # check for no more bugs on json
                                if isinstance(author_name, str):
                                    metadata.author = normalize_authors(metadata.author, author_name)
                            elif 'givenName' in author and 'familyName' in author:
                                name = [author['givenName'], author.get('additionalName'), author['familyName']]
                                metadata.author = normalize_authors(metadata.author, ' '.join(filter(None, name)))

                # category
                if metadata.categories is None and 'articleSection' in content:
                    if isinstance(content['articleSection'], str):
                        metadata.categories = [content['articleSection']]
                    else:
                        metadata.categories = list(filter(None, content['articleSection']))

                # try to extract title
                if metadata.title is None:
                    if 'name' in content and content_type == 'article':
                        metadata.title = content['name']
                    elif 'headline' in content:
                        metadata.title = content['headline']

    return metadata


def extract_json_author(elemtext, regular_expression):
    '''Crudely extract author names from JSON-LD data'''
    authors = None
    mymatch = regular_expression.search(elemtext)
    while mymatch is not None and mymatch[1] and ' ' in mymatch[1]:
        authors = normalize_authors(authors, mymatch[1])
        elemtext = regular_expression.sub(r'', elemtext, count=1)
        mymatch = regular_expression.search(elemtext)
    return authors or None


def extract_json_parse_error(elem, metadata):
    '''Crudely extract metadata from JSON-LD data'''
    # author info
    element_text_author = JSON_AUTHOR_REMOVE.sub('', elem)
    if any(JSON_MATCH.findall(element_text_author)):
        author = extract_json_author(element_text_author, JSON_AUTHOR_1) or extract_json_author(element_text_author, JSON_AUTHOR_2)
        if author:
            metadata.author = author

    # try to extract page type as an alternative to og:type
    if "@type" in elem:
        mymatch = JSON_TYPE.search(elem)
        candidate = normalize_json(mymatch[1].lower())
        if mymatch and candidate in JSON_OGTYPE_SCHEMA:
            metadata.pagetype = candidate

    # try to extract publisher
    if '"publisher"' in elem:
        mymatch = JSON_PUBLISHER.search(elem)
        if mymatch and ',' not in mymatch[1]:
            candidate = normalize_json(mymatch[1])
            if metadata.sitename is None or len(metadata.sitename) < len(candidate):
                metadata.sitename = candidate
            if metadata.sitename.startswith('http') and not candidate.startswith('http'):
                metadata.sitename = candidate

    # category
    if '"articleSection"' in elem:
        mymatch = JSON_CATEGORY.search(elem)
        if mymatch:
            metadata.categories = [normalize_json(mymatch[1])]

    # try to extract title
    if '"name"' in elem and metadata.title is None:
        mymatch = JSON_NAME.search(elem)
        if mymatch:
            metadata.title = normalize_json(mymatch[1])
    if '"headline"' in elem and metadata.title is None:
        mymatch = JSON_HEADLINE.search(elem)
        if mymatch:
            metadata.title = normalize_json(mymatch[1])

    # exit if found
    return metadata


def normalize_json(inputstring):
    'Normalize unicode strings and trim the output'
    if '\\' in inputstring:
        inputstring = inputstring.replace('\\n', '').replace('\\r', '').replace('\\t', '')
        inputstring = JSON_UNICODE_REPLACE.sub(lambda match: chr(int(match.group(1), 16)), inputstring)
        inputstring = ''.join(c for c in inputstring if ord(c) < 0xD800 or ord(c) > 0xDFFF)
        inputstring = unescape(inputstring)
    return trim(JSON_REMOVE_HTML.sub('', inputstring))



import re
from html import unescape

import cchardet as chardet
import lxml
from charset_normalizer import from_bytes
from resiliparse.parse.html import DOMCollection, HTMLTree
from tabulate import tabulate

from haruka_parser.line_processing import (have_chinese_characters,
                                           restore_replacements)
from haruka_parser.time_formatter import return_format_datetime
from haruka_parser.utils import has_style, extract_code_languages


UNICODE_ALIASES = {"utf-8", "utf_8"}

header_to_format = {f"h{i}": f"[heading_{i}]" for i in range(1, 7)}

# Step 2: Use regular expressions to find charset declarations
charset_pattern = re.compile(r'charset=["\']?([\w-]+)["\']?', re.IGNORECASE)
xml_pattern = re.compile(r'encoding=["\']?([\w-]+)["\']?', re.IGNORECASE)

FIX_CHARSETS = {
    "big5": "big5hkscs",
    "gb2312": "gb18030",
    "ascii": "utf-8",
    "maccyrillic": "cp1251",
    "win1251": "cp1251",
    "win-1251": "cp1251",
    "windows-1251": "cp1251",
    "latin": "ISO-8859-1",
    "latin1": "ISO-8859-1",
    "latin-1": "ISO-8859-1",
    "cp1252": "windows-1252",
    "windows-1252": "windows-1252",
}


def fix_charset(encoding):
    """Overrides encoding when charset declaration
    or charset determination is a subset of a larger
    charset.  Created because of issues with Chinese websites"""
    encoding = encoding.lower()
    return FIX_CHARSETS.get(encoding, encoding)


def fix_encode_html(raw_html, encoding="utf-8"):
    # encode raw html in the correct encoding

    try:
        return raw_html.decode('UTF-8'), 'utf-8'
    except UnicodeDecodeError:
        pass

    black_detect_list = set(["windows-1252"])
    guesses = [
        # Japanese
        "shift_jis",
        "euc-jp",
        "iso2022_jp",
        # Simplified/Traditional Chinese
        "big5hkscs",
        "gb18030",
        # Korean
        "iso2022_kr",
        # Cyrillic
        "koi8-r",
        "koi8-u",
        # Alternatives for Europe
        "windows-1252",
        "iso-8859-2",
        "iso-8859-5",
        "iso-8859-7",
        "iso-8859-9",
    ]

    def add_to_guesses(encoding):
        if fix_charset(encoding) not in black_detect_list:
            guesses.insert(0, encoding)

    if len(raw_html) < 10000:
        detection_results = from_bytes(raw_html)
    else:
        detection_results = from_bytes(
            raw_html[:5000] + raw_html[-5000:]
        ) or from_bytes(raw_html)
        
    for r in detection_results:
        add_to_guesses(r.encoding.lower())

    encode_type = chardet.detect(raw_html).get("encoding")
    if encode_type is not None:
        add_to_guesses(encode_type.lower())

    decoded_html = raw_html[:1000].decode("utf-8", errors="ignore")
    charsets = charset_pattern.findall(decoded_html) + xml_pattern.findall(decoded_html)
    for encode_type in charsets:
        add_to_guesses(encode_type.lower())

    add_to_guesses(encoding.lower())

    guesses_done = set()

    for guess_type in guesses:
        fix_guess_type = fix_charset(guess_type)
        if fix_guess_type not in UNICODE_ALIASES and fix_guess_type not in guesses_done:
            guesses_done.add(fix_guess_type)
            try:
                content = raw_html.decode(fix_guess_type)
                return content, fix_guess_type
            except:
                pass

    return None, None

def remove_jax_ignore(tree):
    # select all *jax_ignore class
    elements = tree.document.query_selector_all("[class*='jax_ignore']")
    for element in elements:
        parent = element.parent
        if parent:
            parent.remove_child(element)


def remove_buttons(tree):
    btns = tree.document.query_selector_all(".btn")
    for btn in btns:
        parent = btn.parent
        parent.remove_child(btn)
    # Remove any button tags
    btns = tree.document.query_selector_all("button")
    for btn in btns:
        parent = btn.parent
        if parent:
            parent.remove_child(btn)
    # Remove any element with onclick attr
    btns = tree.document.query_selector_all("[onclick]")
    for btn in btns:
        parent = btn.parent
        if parent:
            parent.remove_child(btn)
    btns = tree.document.query_selector_all("a[class*='down']")
    for btn in btns:
        parent = btn.parent
        if parent:
            parent.remove_child(btn)


def remove_links(tree):
    """Replace links with spans so that resiliparse doesn't try to remove them."""
    links = tree.document.query_selector_all("a")
    for link in links:
        parent = link.parent
        if parent is None:
            continue
        new_span = tree.create_element("span")
        new_span.text = link.text
        parent.replace_child(new_span, link)


def flatten(node):
    """Remove any divs or spans that only have one child and replace them with their child."""
    divs = node.query_selector_all("div")
    spans = node.query_selector_all("span")
    for div in divs:
        if len(div.child_nodes) == 1:
            parent = div.parent
            if parent is None:
                continue
            parent.replace_child(div.child_nodes[0], div)
    for span in spans:
        if len(span.child_nodes) == 1:
            parent = span.parent
            if parent is None:
                continue
            parent.replace_child(span.child_nodes[0], span)

    return node


def remove_dense_links(tree):
    """Remove lists that only have links."""
    # First, remove any nav elements to be safe.
    navs = tree.document.query_selector_all("nav")
    for nav in navs:
        parent = nav.parent
        if parent is None:
            continue
        parent.remove_child(nav)

    lists = tree.document.query_selector_all("ul, ol, div, span, nav, table, p")
    to_remove = []
    for _list in lists:
        if len(_list.child_nodes) == 0 or len(_list.child_nodes) == 1:
            continue
        children = _list.child_nodes
        links = _list.query_selector_all("a")
        total_children_text = "".join(
            [x.text.strip() for x in children if type(x) != DOMCollection]
        )
        total_links_text = "".join([x.text.strip() for x in links])
        if len(total_children_text) == 0 or len(total_links_text) == 0:
            continue
        ratio = len(total_links_text) / len(total_children_text)
        if ratio > 0.8:
            parent = _list.parent
            if parent is None:
                continue
            to_remove.append(_list)

    for _list in to_remove:
        parent = _list.parent
        if parent is None:
            continue
        parent.remove_child(_list)


def remove_image_figures(tree):
    to_remove = []
    imgs = tree.document.query_selector_all("img")
    for img in imgs:
        cur_node = img
        while cur_node is not None:
            if cur_node.class_name == "figure":
                parent = cur_node.parent
                if parent:
                    to_remove.append(cur_node)
                break
            cur_node = cur_node.parent

    for node in to_remove:
        parent = node.parent
        if parent is None:
            continue
        parent.remove_child(node)


def remove_link_clusters(tree):
    # First, find all links that are in span blocks. If they have no siblings, delete the span.
    to_remove = []

    span_links = tree.document.query_selector_all("span a")
    for link in span_links:
        parent = link.parent
        if parent is None:
            continue
        n_siblings = 0
        for sibling in parent.child_nodes:
            if sibling.type == 1:
                n_siblings += 1
                break
        if n_siblings == 1:
            grandparent = parent.parent
            if grandparent is None:
                continue
            # grandparent.remove_child(parent)
            to_remove.append(parent)

    links = list(tree.document.query_selector_all("a"))

    i = 0
    while len(links) > 0:
        link = links[0]
        del links[0]
        parent = link.parent
        i += 1
        if parent is None or parent.parent is None:
            continue
        n_links = 0
        n_children = len(parent.child_nodes)
        child_links = parent.query_selector_all("a")
        if len(child_links) == n_children:
            for child_link in child_links:
                # Check if it's visible and not empty.
                empty = child_link.text is None or child_link.text.strip() == ""
                styles = child_link.getattr("style")
                visible = styles is None or not (
                    has_style("display: none", styles)
                    or has_style("visibility: hidden", styles)
                )
                if visible and not empty:
                    n_links += 1
            multilink = n_links > 1 and n_children == n_links
            if multilink:
                grandparent = parent.parent
                if grandparent is None:
                    continue
                # grandparent.remove_child(parent)
                to_remove.append(parent)

    for node in to_remove:
        parent = node.parent
        if parent is None:
            continue
        parent.remove_child(node)


def extract_code(tree, replacement_manager, info, enable_code=False):
    wp_syntax = tree.document.query_selector_all(".wp_syntax")
    codes = tree.document.query_selector_all("code")
    code_responsive = tree.document.query_selector_all(".code_responsive")
    pre_tags = []
    # current disable pre tag due to false positive
    # pre_tags = tree.document.query_selector_all("pre")
    for code in [*wp_syntax, *codes, *code_responsive, *pre_tags]:
        if replacement_manager.add_replacement("", tag="code") in code.text:
            continue
        multiline = code.text.strip().count("\n") > 0

        cur_code_language = extract_code_languages(code.class_name)
        info["code_language"] += cur_code_language

        if cur_code_language:
            cur_code_language_text = cur_code_language[0]

        if len(code.text) > 0:
            info["code_block"] += 1
            if enable_code:
                if multiline:
                    code_text = replacement_manager.add_replacement(
                        f"[three_code_dot]{cur_code_language_text}\n{code.text.strip()}\n[three_code_dot]",
                        tag="code",
                    )
                else:
                    code_text = replacement_manager.add_replacement(
                        f"[single_code_dot]{code.text}[single_code_dot]", tag="code"
                    )
                code_text = code_text.replace(" ", "[HARUKA_PARSER_BS]")

                new_span = tree.create_element("span")
                new_span.text = code_text
                code.parent.replace_child(new_span, code)

    return info


def extract_tables(node, replacement_manager, config, info):
    # Don't worry about tables that have tables in them or have headers
    # tables = node.query_selector_all('table:not(:has(table *))')
    tables = node.query_selector_all("table:not(:has(table))")
    for table in tables:
        # restore table first in order to handle indent corrently
        table.html = restore_replacements(table.html, replacement_manager, config, info)
    # if config["table_config"]["format"] == "none":
    #     return
    tables = node.query_selector_all("table:not(:has(table, h1, h2, h3, h4, h5, h6))")
    for table in tables:
        table_data = []
        headers = []
        # Find all headers
        ths = table.query_selector_all("th")
        for th in ths:
            headers.append(th.text)
        trs = table.query_selector_all("tr")
        for tr in trs:
            row_data = []
            tds = tr.query_selector_all("td")
            for td in tds:
                # Remove any scripts
                scripts = td.query_selector_all("script")
                for script in scripts:
                    script.parent.remove_child(script)
                # Get the text of each td element
                row_data.append(td.text)
                col_span = td.getattr("colspan")
                if col_span:
                    try:
                        col_span = int(col_span)
                        if col_span > 100:
                            continue
                    except ValueError:
                        continue
                    # Add empty cells for colspans
                    for _ in range(col_span - 1):
                        row_data.append("")
            if row_data:
                table_data.append(row_data)
        if len(table_data) == 0 or len(table_data[0]) == 0:
            continue
        # Post processing
        # Make sure all rows have the same number of columns
        max_cols = max([len(row) for row in table_data])
        for row in table_data:
            if len(row) < max_cols:
                row.extend([""] * (max_cols - len(row)))
        # Strip all cells
        for i in range(len(table_data)):
            for j in range(len(table_data[i])):
                table_data[i][j] = table_data[i][j].strip()
        # If any columns or rows are consistently empty, remove them
        # Remove empty columns
        empty_columns = []
        for i in range(len(table_data[0])):
            if all([len(row[i]) == 0 for row in table_data]):
                empty_columns.append(i)

        for i in reversed(empty_columns):
            for row in table_data:
                del row[i]
        # Remove empty rows
        table_data = [row for row in table_data if len(row) > 0]

        # Remove any newlines from the table
        for i in range(len(table_data)):
            for j in range(len(table_data[i])):
                table_data[i][j] = table_data[i][j].replace("\n", " ")
        # Check that the table has at least one row and one column
        # if (
        #     len(table_data) >= config["table_config"]["min_rows"]
        #     and len(table_data[0]) >= config["table_config"]["min_cols"]
        # ):
        if len(table_data) >= 1 and len(table_data[0]) >= 1:
            # Replace the table with a markdown
            parent = table.parent
            if parent:
                if len(headers) == 0:
                    headers = [""] * len(table_data[0])
                rendered_table = tabulate(
                    table_data,
                    # tablefmt=config["table_config"]["format"],
                    tablefmt="html",
                    headers=headers,
                )
                info["table"] += 1
                if have_chinese_characters(rendered_table):
                    info["chinese_table"] += 1
                table.html = replacement_manager.add_replacement(
                    rendered_table, tag="table"
                )
        elif len(table_data) > 0 and len(table_data[0]) > 0:
            # Do the same but use a plain format
            # Replace the table with a markdown
            parent = table.parent
            if parent:
                if len(headers) == 0:
                    headers = [""] * len(table_data[0])
                rendered_table = tabulate(table_data, tablefmt="plain", headers=headers)
                table.html = replacement_manager.add_replacement(
                    rendered_table, tag="table"
                )
        else:
            # Remove empty tables
            if table.parent:
                table.parent.remove_child(table)


def extract_headings(tree, replacement_manager, markdown_formatting):
    to_remove = []
    for heading_tag in header_to_format:
        hs = tree.document.query_selector_all(heading_tag)
        for heading in hs:
            text = ""
            for child in heading.child_nodes:
                if child.text.strip() != "" and child.type != 8:
                    text += child.text
                    child.text = ""
            text = text.strip()
            if len(text) == 0:
                # remove the heading
                if heading.parent:
                    to_remove.append(heading)
                continue
            if markdown_formatting:
                heading.text = replacement_manager.add_replacement(
                    header_to_format[heading_tag] + " " + text + "\n\n", tag=heading_tag
                )
            else:
                heading.text = replacement_manager.add_replacement(
                    text + "\n\n", tag=heading_tag
                )

    for heading in to_remove:
        parent = heading.parent
        if parent:
            parent.remove_child(heading)


def post_process_headings(text):
    """Replace [heading_i] with '#' * i"""
    for i in range(6, 0, -1):
        text = text.replace("[heading_%d]" % i, "#" * i)
    return text


def add_se_separators(tree):
    user_infos = tree.document.query_selector_all("table.fw")
    # Replace all of these with spans <span>-</span>
    for user_info in user_infos:
        new_span = tree.create_element("span")
        new_span.text = "-"
        parent = user_info.parent
        # Remove the table
        parent.remove_child(user_info)
        # Add the span
        parent.append_child(new_span)


def wikipedia_preprocess(tree):
    external_links = tree.document.query_selector("#External_links")
    if external_links:
        # Remove all next until nothing left
        node = external_links.parent.next
        while node:
            next = node.next
            node.parent.remove_child(node)
            node = next
        external_links.parent.remove_child(external_links)

    edit_buttons = tree.document.query_selector_all(".mw-editsection")
    for edit_button in edit_buttons:
        if edit_button.parent:
            edit_button.parent.remove_child(edit_button)


def remove_display_none(tree):
    # Remove all elements with display none
    elements = tree.document.query_selector_all('[style*="display:none"]')
    # elements = tree.document.query_selector_all('[style*="display:none"], [style*="display: none"]')
    for element in elements:
        element.parent.remove_child(element)

def remove_svg(tree):
    # Remove all elements with display none
    elements = tree.document.query_selector_all('svg')
    for element in elements:
        element.parent.remove_child(element)


def preserve_question_headers(tree):
    elements = tree.document.query_selector_all("#question-header")
    for element in elements:
        inner_h1 = element.query_selector("h1")
        if inner_h1:
            new_h1 = tree.create_element("h1")
            new_h1.text = inner_h1.text
            element.parent.replace_child(new_h1, element)


def main_content_preprocess(tree):
    """Make any changes that are necessary to maximize the performance
    of the resiliparse main_content=True option."""

    # Look for qa-main class
    qa_main = tree.document.query_selector(".qa-main")
    if qa_main:
        qa_main.setattr("class", "article-body")

    # If there is a role=main and a question-header class, add the question-header to the top of the role=main
    role_main = tree.document.query_selector('[role="main"]')
    if role_main:
        question_header = tree.document.query_selector("#question-header")
        if question_header:
            first_child = role_main.first_child
            if first_child:
                role_main.insert_before(question_header, first_child)

    post_content = tree.document.query_selector(".postcontent")
    if post_content:
        post_body = tree.document.query_selector(".postbody")
        if post_body:
            # Set the class of postbody to postcontent and remove the postcontent class
            post_body.setattr("class", "postcontent")
            post_content.setattr("class", "")

    # Find .postbit
    postbit = tree.document.query_selector(".postbit")
    if postbit:
        # Change the class to article-body
        postbit.setattr("class", "")

    # Find all ul and add a few wrapping divs to move them farther from the root node
    uls = tree.document.query_selector_all("ul")
    for ul in uls:
        # Create 4 nested divs and set the html of the last one to the html of the ul. Then replace the ul with the last div
        div1 = tree.create_element("div")
        div2 = tree.create_element("div")
        div3 = tree.create_element("div")
        div4 = tree.create_element("div")
        div4.html = ul.html
        div3.append_child(div4)
        div2.append_child(div3)
        div1.append_child(div2)
        if ul.parent:
            ul.parent.replace_child(div1, ul)


RE_STRIP_XML_DECLARATION = re.compile(r"^<\?xml [^>]+?\?>")


def normalize_spaces(s):
    if not s:
        return ""
    """replace any sequence of whitespace
    characters with a single space"""
    return " ".join(s.split())


def normalize_entities(cur_title):
    entities = {
        "\u2014": "-",
        "\u2013": "-",
        "&mdash;": "-",
        "&ndash;": "-",
        "\u00A0": " ",
        "\u00AB": '"',
        "\u00BB": '"',
        "&quot;": '"',
    }
    for c, r in entities.items():
        if c in cur_title:
            cur_title = cur_title.replace(c, r)

    return cur_title


def norm_title(title):
    return normalize_entities(normalize_spaces(title))


def add_match(collection, text, orig):
    text = norm_title(text)
    if (len(text.split()) >= 2 and len(text) >= 15) or (
        have_chinese_characters(text) and len(text) >= 5
    ):
        if text.replace('"', "") in orig.replace('"', ""):
            collection.add(text)


META_TITLE_XPATHS = [
    'meta[property="og:title"]',
    'meta[name="twitter:title"]',
    'meta[property="twitter:title"]',
]

# order matters
TITLE_CSS_HEURISTICS = [
    "h1",
    "h2",
    "h3",
    "#title",
    "#Title",
    "#TITLE",
    ".title",
    ".Title",
    ".TITLE",
    "#head",
    ".head",
    "#heading",
    ".heading",
    ".pageTitle",
    ".news_title",
    ".contentheading",
    ".small_header_red",
]


def _get_title(doc: HTMLTree) -> str:
    # parse twitter
    for item in META_TITLE_XPATHS:
        for e in doc.body.query_selector_all(item):
            if e.getattr("content"):
                return e.getattr("content")

    if doc is None:
        return ""
    title = doc.title.strip()
    if title:
        title = orig = norm_title(title)
    else:
        orig = ""

    candidates = set()

    for item in TITLE_CSS_HEURISTICS:
        for e in doc.body.query_selector_all(item):
            if e.text:
                if not orig:
                    orig = e.text
                else:
                    add_match(candidates, e.text, orig)

    if candidates:
        title = sorted(candidates, key=len)[-1]
    else:
        for delimiter in [" | ", " - ", " :: ", " / "]:
            if delimiter in title:
                parts = orig.split(delimiter)
                if len(parts[0].split()) >= 4:
                    title = parts[0]
                    break
                elif len(parts[-1].split()) >= 4:
                    title = parts[-1]
                    break
        else:
            if ": " in title:
                parts = orig.split(": ")
                if len(parts[-1].split()) >= 4:
                    title = parts[-1]
                else:
                    title = orig.split(": ", 1)[1]

    if not 5 < len(title) < 150:
        return orig or ""

    return title or ""


def get_title(doc: HTMLTree) -> str:
    return unescape(_get_title(doc))


def get_lxml_tree(html_content):
    """Obtain the HTML parse tree for the given HTML content.

    Args:
        html_content: The content to parse.

    Returns:
        The corresponding HTML parse tree.
    """
    html_content = html_content.strip()
    if not html_content:
        return None

    # strip XML declaration, if necessary
    if html_content.startswith("<?xml "):
        html_content = RE_STRIP_XML_DECLARATION.sub("", html_content, count=1)

    try:
        return lxml.html.document_fromstring(
            html_content.encode("utf-8", "replace"),
            parser=lxml.html.HTMLParser(encoding="utf-8"),
        )
    except:
        try:
            return lxml.html.fromstring("<pre>" + html_content + "</pre>")
        except:
            return None


META_TIME_XPATHS = [
    '//meta[contains(@name, "og:time")]/@content',
    '//meta[contains(@name, "PubDate")]/@content',
    '//meta[contains(@name, "pubtime")]/@content',
    '//meta[contains(@name, "_pubtime")]/@content',
    '//meta[contains(@name, "apub:time")]/@content',
    '//meta[contains(@pubdate, "pubdate")]/@content',
    '//meta[contains(@name, "publishdate")]/@content',
    '//meta[contains(@name, "PublishDate")]/@content',
    '//meta[contains(@name, "sailthru.date")]/@content',
    '//meta[contains(@itemprop, "dateUpdate")]/@content',
    '//meta[contains(@name, "publication_date")]/@content',
    '//meta[contains(@itemprop, "datePublished")]/@content',
    '//meta[contains(@property, "og:release_date")]/@content',
    '//meta[contains(@name, "article_date_original")]/@content',
    '//meta[contains(@property, "og:published_time")]/@content',
    '//meta[contains(@property, "rnews:datePublished")]/@content',
    '//meta[contains(@name, "OriginalPublicationDate")]/@content',
    '//meta[contains(@name, "weibo: article:create_at")]/@content',
    '//meta[@name="Keywords" and contains(@content, ":")]/@content',
    '//meta[contains(@property, "article:published_time")]/@content',
]

SUPPLEMENT_TIME_XPATHS = [
    '//div[@class="time fix"]//text()',
    '//span[@id="pubtime_baidu"]/text()',
    '//i[contains(@class, "time")]/text()',
    '//span[contains(text(), "时间")]/text()',
    '//div[contains(@class, "time")]//text()',
    '//span[contains(@class, "date")]/text()',
    '//div[contains(@class, "info")]//text()',
    '//span[contains(@class, "time")]/text()',
    '//div[contains(@class, "_time")]/text()',
    '//span[contains(@id, "paperdate")]/text()',
    '//em[contains(@id, "publish_time")]/text()',
    '//time[@data-testid="timestamp"]/@dateTime',
    '//span[contains(@id, "articleTime")]/text()',
    '//span[contains(@class, "pub_time")]/text()',
    '//span[contains(@class, "item-time")]/text()',
    '//span[contains(@class, "publishtime")]/text()',
    '//div[contains(@class, "news_time_source")]/text()',
]

REGEX_TIME = [
    r"(\d{1,2}月\d{1,2}日)",
    r"(\d{2}年\d{1,2}月\d{1,2}日)",
    r"(\d{4}年\d{1,2}月\d{1,2}日)",
    r"(\d{2}[-|/|.]\d{1,2}[-|/|.]\d{1,2})",
    r"(\d{4}[-|/|.]\d{1,2}[-|/|.]\d{1,2})",
    r"(\d{1,2}月\d{1,2}日\s*?[2][0-3]:[0-5]?[0-9])",
    r"(\d{1,2}月\d{1,2}日\s*?[0-1]?[0-9]:[0-5]?[0-9])",
    r"(\d{2}年\d{1,2}月\d{1,2}日\s*?[2][0-3]:[0-5]?[0-9])",
    r"(\d{4}年\d{1,2}月\d{1,2}日\s*?[2][0-3]:[0-5]?[0-9])",
    r"(\d{2}年\d{1,2}月\d{1,2}日\s*?[0-1]?[0-9]:[0-5]?[0-9])",
    r"(\d{4}年\d{1,2}月\d{1,2}日\s*?[0-1]?[0-9]:[0-5]?[0-9])",
    r"(\d{1,2}月\d{1,2}日\s*?[1-24]\d时[0-60]\d分)([1-24]\d时)",
    r"(\d{1,2}月\d{1,2}日\s*?[2][0-3]:[0-5]?[0-9]:[0-5]?[0-9])",
    r"(\d{4}[-|/|.]\d{1,2}[-|/|.]\d{1,2}\s*?[2][0-3]:[0-5]?[0-9])",
    r"(\d{1,2}月\d{1,2}日\s*?[0-1]?[0-9]:[0-5]?[0-9]:[0-5]?[0-9])",
    r"(\d{2}[-|/|.]\d{1,2}[-|/|.]\d{1,2}\s*?[2][0-3]:[0-5]?[0-9])",
    r"(\d{4}[-|/|.]\d{1,2}[-|/|.]\d{1,2}\s*?[0-1]?[0-9]:[0-5]?[0-9])",
    r"(\d{2}[-|/|.]\d{1,2}[-|/|.]\d{1,2}\s*?[0-1]?[0-9]:[0-5]?[0-9])",
    r"(\d{2}年\d{1,2}月\d{1,2}日\s*?[1-24]\d时[0-60]\d分)([1-24]\d时)",
    r"(\d{4}年\d{1,2}月\d{1,2}日\s*?[1-24]\d时[0-60]\d分)([1-24]\d时)",
    r"(\d{2}年\d{1,2}月\d{1,2}日\s*?[2][0-3]:[0-5]?[0-9]:[0-5]?[0-9])",
    r"(\d{4}年\d{1,2}月\d{1,2}日\s*?[2][0-3]:[0-5]?[0-9]:[0-5]?[0-9])",
    r"(\d{2}年\d{1,2}月\d{1,2}日\s*?[0-1]?[0-9]:[0-5]?[0-9]:[0-5]?[0-9])",
    r"(\d{4}年\d{1,2}月\d{1,2}日\s*?[0-1]?[0-9]:[0-5]?[0-9]:[0-5]?[0-9])",
    r"(\d{4}[-|/|.]\d{1,2}[-|/|.]\d{1,2}\s*?[1-24]\d时[0-60]\d分)([1-24]\d时)",
    r"(\d{2}[-|/|.]\d{1,2}[-|/|.]\d{1,2}\s*?[1-24]\d时[0-60]\d分)([1-24]\d时)",
    r"(\d{2}[-|/|.]\d{1,2}[-|/|.]\d{1,2}\s*?[2][0-3]:[0-5]?[0-9]:[0-5]?[0-9])",
    r"(\d{4}[-|/|.]\d{1,2}[-|/|.]\d{1,2}\s*?[2][0-3]:[0-5]?[0-9]:[0-5]?[0-9])",
    r"(\d{4}[-|/|.]\d{1,2}[-|/|.]\d{1,2}\s*?[0-1]?[0-9]:[0-5]?[0-9]:[0-5]?[0-9])",
    r"(\d{2}[-|/|.]\d{1,2}[-|/|.]\d{1,2}\s*?[0-1]?[0-9]:[0-5]?[0-9]:[0-5]?[0-9])",
]


def get_valid_length_time(publish_time_list):
    if len(publish_time_list) > 0:
        length_valid_publish_time = [y for y in publish_time_list if len(y) >= 9]
        publish_time = (
            length_valid_publish_time[0]
            if length_valid_publish_time
            else publish_time_list[0]
        )
    else:
        publish_time = ""
    return publish_time


def extract_time_from_meta(html_tree):
    for each_meta_xpath in META_TIME_XPATHS:
        publish_time_temp = html_tree.xpath(each_meta_xpath)
        if publish_time_temp:
            return "".join(publish_time_temp).strip()
    return ""


def extract_time_from_other_tag(html_tree):
    publish_time_list = []
    for each_xpath in SUPPLEMENT_TIME_XPATHS:
        publish_time_temp = "".join(html_tree.xpath(each_xpath)).strip()
        if publish_time_temp:
            publish_time_list += [
                re.findall(x, publish_time_temp)[0]
                for x in REGEX_TIME
                if re.findall(x, publish_time_temp)
            ]
    return get_valid_length_time(publish_time_list)


def extract_time_from_html(content):
    publish_time_list = []
    for each_time_regex in REGEX_TIME:
        publish_time_temp = re.findall(each_time_regex, content)
        if publish_time_temp:
            publish_time_list += publish_time_temp
    return get_valid_length_time(publish_time_list)


def extract_time(content, html_tree):
    publish_time = ""
    meta_date = extract_time_from_meta(html_tree)
    if meta_date:
        publish_time = return_format_datetime(meta_date)
    else:
        tag_date = extract_time_from_other_tag(html_tree)
        if tag_date:
            publish_time = return_format_datetime(tag_date)
        else:
            html_date = extract_time_from_html(content)
            if html_date:
                publish_time = return_format_datetime(html_date)
    return publish_time



"Code parts dedicated to duplicate removal and text similarity."


import re
import string

from difflib import SequenceMatcher
from functools import lru_cache
from hashlib import blake2b
from operator import add
from threading import RLock
from typing import Any, List, Optional

from .settings import LRU_SIZE
from .utils import trim


STRIP_EXTENSION = re.compile(r"\.[^/?#]{2,63}$")


@lru_cache(maxsize=1024)
def is_similar_domain(reference: str, new_string: str, threshold: float = 0.5) -> bool:
    "Return the similarity ratio between two short strings, here domain names."
    if new_string != reference:
        new_string = STRIP_EXTENSION.sub("", new_string)
        reference = STRIP_EXTENSION.sub("", reference)
        if SequenceMatcher(None, reference, new_string).ratio() < threshold:
            return False
    return True


def sample_tokens(inputstring: str, length: int = 64) -> List[str]:
    """Split input into list of tokens and adjust length threshold to make sure
    there is enough data."""
    tokens = []
    for token in inputstring.split():
        token = token.strip(string.punctuation)
        if token.isalnum():
            tokens.append(token)
    sample = []
    for i in range(4, -1, -1):
        sample = [t for t in tokens if len(t) > i]
        if len(sample) >= length / 2:
            return sample
    return sample


def generate_bow_hash(inputstring: str, length: int = 24) -> bytes:
    "Create a bag of words and generate a hash for a given string."
    teststring = " ".join(sample_tokens(inputstring)).strip()
    # perform hashing with limited size
    return blake2b(teststring.encode(), digest_size=length).digest()


class Simhash:
    "Implement a basic Charikar hashing approach of string similarity."
    __slots__ = ["hash", "length"]

    def __init__(
        self,
        inputstring: str = "",
        length: int = 64,
        existing_hash: Optional[str] = None,
    ) -> None:
        "Store length and existing or new hash."
        self.length = length
        self.hash = self.validate(existing_hash) or self.create_hash(inputstring)

    def _hash(self, inputstring: str) -> int:
        "Return a numerical hash of the string."
        return int.from_bytes(
            blake2b(inputstring.encode(), digest_size=8).digest(), "big"
        )
        # old: variable-length version of Python's builtin hash by @sean-public
        # see also Siphash13 in https://peps.python.org/pep-0456/
        # if inputstring == "":
        #    return 0
        # mask = 2**self.length - 1
        # x = ord(inputstring[0]) << 7
        # for c in inputstring:
        #    x = ((x * 1000003) ^ ord(c)) & mask
        # x ^= len(inputstring)
        # if x == -1:
        #    return -2
        # return x

    @lru_cache(maxsize=2**14)
    def _vector_to_add(self, token: str) -> List[int]:
        "Create vector to add to the existing string vector"
        the_hash = self._hash(token)
        return [1 if the_hash & (1 << i) else -1 for i in range(self.length)]

    def create_hash(self, inputstring: str) -> int:
        """Calculates a Charikar simhash. References used:
        https://github.com/vilda/shash/
        https://github.com/sean-public/python-hashes/blob/master/hashes/simhash.py
        Optimized for Python by @adbar.
        """
        vector = [0] * self.length

        for token in sample_tokens(inputstring, self.length):
            vector = list(map(add, vector, self._vector_to_add(token)))

        return sum(1 << i for i in range(self.length) if vector[i] >= 0)

    def to_hex(self) -> str:
        "Convert the numerical hash to a hexadecimal string."
        return hex(self.hash)[2:]

    def _hash_to_int(self, inputhash) -> Optional[int]:
        "Convert the hexadecimal hash to a numerical value."
        try:
            return int(inputhash, 16)
        except (TypeError, ValueError):
            return None

    def validate(self, inputhash: Optional[Any]) -> Optional[int]:
        "Validate the input hash and return it, or None otherwise."
        if isinstance(inputhash, int) and 18 <= len(str(inputhash)) <= 22:
            return inputhash
        if isinstance(inputhash, str):
            if inputhash.isdigit() and 18 <= len(inputhash) <= 22:
                return int(inputhash)
            # possibly a hex string
            return self._hash_to_int(inputhash)
        return None

    def hamming_distance(self, other_hash: Any) -> int:
        "Return distance between two hashes of equal length using the XOR operator."
        xor_result = self.hash ^ other_hash.hash
        try:
            # Python >= 3.10
            return xor_result.bit_count()
        except AttributeError:
            return bin(xor_result).count("1")

    def similarity(self, other_hash: Any) -> float:
        """Calculate how similar this hash is from another simhash.
        Returns a float from 0.0 to 1.0.
        """
        return (self.length - self.hamming_distance(other_hash)) / self.length


def content_fingerprint(content: str) -> str:
    "Calculate a simhash hex value for meaningful bits of the content."
    return Simhash(content).to_hex()


PREV, NEXT, KEY, RESULT = 0, 1, 2, 3  # names for the link fields


class LRUCache:
    """
    Pure-Python Least Recently Used (LRU) cache using a circular doubly linked list
    Adapted from CPython functools.py lru_cache decorator implementation
    https://github.com/python/cpython/blob/3.9/Lib/functools.py#L524
    First adapted by https://github.com/vbarbaresi
    """

    def __init__(self, maxsize: int = 128) -> None:
        # Constants shared by all lru cache instances:
        self.lock = RLock()  # because linkedlist updates aren't threadsafe
        # cache instance variables
        self.maxsize = maxsize
        self.cache = {}
        self.root = []  # root of the circular doubly linked list
        # initialize by pointing to self
        self.root[:] = [self.root, self.root, None, None]
        self.full = False

    def _move_link(self, link: Any) -> Any:
        # Move the link to the front of the circular queue
        link_prev, link_next, _key, result = link
        link_prev[NEXT], link_next[PREV] = link_next, link_prev
        last = self.root[PREV]
        last[NEXT] = self.root[PREV] = link
        link[PREV] = last
        link[NEXT] = self.root
        return result

    def get(self, key) -> Any:
        """Tests if the key that is asked for is in the cache
           and retrieve its value from the linked list."""
        with self.lock:
            link = self.cache.get(key)
            if link:
                return self._move_link(link)
        return -1

    def put(self, key, value) -> None:
        "Stores a given key in the cache."
        # Size limited caching that tracks accesses by recency
        with self.lock:
            link = self.cache.get(key)
            if link:
                self._move_link(link)
                self.cache[key][RESULT] = value
            else:
                if self.full:
                    # Use the old root to store the new key and result.
                    oldroot = self.root
                    oldroot[KEY], oldroot[RESULT] = key, value
                    # Empty the oldest link and make it the new root.
                    # Keep a reference to the old key and old result to
                    # prevent their ref counts from going to zero during the
                    # update. That will prevent potentially arbitrary object
                    # clean-up code (i.e. __del__) from running while we're
                    # still adjusting the links.
                    self.root = oldroot[NEXT]
                    oldkey = self.root[KEY]
                    self.root[KEY] = self.root[RESULT] = None
                    # Now update the cache dictionary.
                    del self.cache[oldkey]
                    # Save the potentially reentrant cache[key] assignment
                    # for last, after the root and links have been put in
                    # a consistent state.
                    self.cache[key] = oldroot
                else:
                    # Put result in a new link at the front of the queue.
                    last = self.root[PREV]
                    link = [last, self.root, key, value]
                    last[NEXT] = self.root[PREV] = self.cache[key] = link
                    # Use the cache_len bound method instead of the len() function
                    # which could potentially be wrapped in an lru_cache itself.
                    self.full = len(self.cache) >= self.maxsize

    def clear(self) -> None:
        "Delete all cache content."
        with self.lock:
            self.cache.clear()
            self.root[:] = [self.root, self.root, None, None]
            self.full = False


LRU_TEST = LRUCache(maxsize=LRU_SIZE)


def put_in_cache(teststring: str) -> None:
    "Implement LRU cache."
    cacheval = LRU_TEST.get(teststring)
    # if the value is already defined
    value = cacheval + 1 if cacheval != -1 else 1
    LRU_TEST.put(teststring, value)


def duplicate_test(element, options):
    "Check for duplicate text with LRU cache."
    teststring = trim(' '.join(element.itertext()))
    # teststring = element.text
    if len(teststring) > options.min_duplcheck_size:
        # retrieve value from cache
        cacheval = LRU_TEST.get(teststring)
        if cacheval > options.max_repetitions:  # non-existent key will return -1
            LRU_TEST.put(teststring, cacheval + 1)
            return True
    put_in_cache(teststring)
    return False



# pylint:disable-msg=E0611
"""
Listing a series of settings that are applied module-wide.
"""

from configparser import ConfigParser
from datetime import datetime

try:
    from os import sched_getaffinity
except ImportError:
    sched_getaffinity = None
    from os import cpu_count

from pathlib import Path

from lxml.etree import XPath


def use_config(filename=None, config=None):
    """
    Use configuration object or read and parse a settings file.
    """
    if config is not None:
        return config

    if filename is None:
        filename = str(Path(__file__).parent / "settings.cfg")
    elif not Path(filename).is_file():
        raise FileNotFoundError("The given config file does not exist")

    config = ConfigParser()
    config.read(filename)
    return config


DEFAULT_CONFIG = use_config()

CONFIG_MAPPING = {
    'min_extracted_size': 'MIN_EXTRACTED_SIZE',
    'min_output_size': 'MIN_OUTPUT_SIZE',
    'min_output_comm_size': 'MIN_OUTPUT_COMM_SIZE',
    'min_extracted_comm_size': 'MIN_EXTRACTED_COMM_SIZE',
    'min_duplcheck_size': 'MIN_DUPLCHECK_SIZE',
    'max_repetitions': 'MAX_REPETITIONS',
    'max_file_size': 'MAX_FILE_SIZE',
    'min_file_size': 'MIN_FILE_SIZE'
}


class Extractor:
    "Defines a class to store all extraction options."
    __slots__ = [
    'config',
    # general
    'format', 'fast', 'focus', 'comments',
    'formatting', 'links', 'images', 'tables', 'dedup', 'lang',
    # extraction size
    'min_extracted_size', 'min_output_size',
    'min_output_comm_size', 'min_extracted_comm_size',
    # deduplication
    'min_duplcheck_size', 'max_repetitions',
    # rest
    'max_file_size', 'min_file_size', 'max_tree_size',
    # meta
    'source', 'url', 'with_metadata', 'only_with_metadata', 'tei_validation',
    'date_params',
    'author_blacklist', 'url_blacklist'
    ]
    # consider dataclasses for Python 3.7+
    def __init__(self, *, config=DEFAULT_CONFIG, output_format="txt",
                 fast=False, precision=False, recall=False,
                 comments=True, formatting=False, links=False, images=False,
                 tables=True, dedup=False, lang=None, max_tree_size=None,
                 url=None, source=None, with_metadata=False, only_with_metadata=False, tei_validation=False,
                 author_blacklist=None, url_blacklist=None, date_params=None):
        self._add_config(config)
        self.format = output_format
        self.fast = fast
        self.focus = "recall" if recall else "precision" if precision else "balanced"
        self.comments = comments
        self.formatting = formatting or output_format == "markdown"
        self.links = links
        self.images = images
        self.tables = tables
        self.dedup = dedup
        self.lang = lang
        self.max_tree_size = max_tree_size
        self.url = url
        self.source = url or source
        self.only_with_metadata = only_with_metadata
        self.tei_validation = tei_validation
        self.author_blacklist = author_blacklist or set()
        self.url_blacklist = url_blacklist or set()
        self.with_metadata = (with_metadata or only_with_metadata or
                              url_blacklist or output_format == "xmltei")
        self.date_params = (date_params or
                            set_date_params(self.config.getboolean('DEFAULT', 'EXTENSIVE_DATE_SEARCH')))

    def _add_config(self, config):
        "Store options loaded from config file."
        for key, value in CONFIG_MAPPING.items():
            setattr(self, key, config.getint('DEFAULT', value))
        self.config = config


def args_to_extractor(args, url=None):
    "Derive extractor configuration from CLI args."
    options = Extractor(
                  config=use_config(filename=args.config_file), output_format=args.output_format,
                  precision=args.precision, recall=args.recall,
                  comments=args.no_comments, tables=args.no_tables,
                  dedup=args.deduplicate, lang=args.target_language, url=url,
                  with_metadata=args.with_metadata, only_with_metadata=args.only_with_metadata,
                  tei_validation=args.validate_tei
              )
    for attr in ("fast", "formatting", "images", "links"):
        setattr(options, attr, getattr(args, attr))
    return options


def set_date_params(extensive=True):
    "Provide default parameters for date extraction."
    return {
               "original_date": True,
               "extensive_search": extensive,
               "max_date": datetime.now().strftime("%Y-%m-%d")
           }


# Safety checks
PARALLEL_CORES = min(len(sched_getaffinity(0)) if sched_getaffinity else cpu_count(), 16)  # 16 processes at most
LRU_SIZE = 4096

# Files
MAX_FILES_PER_DIRECTORY = 1000
FILENAME_LEN = 8

# Network
MAX_LINKS = 10**6
MAX_SITEMAPS_SEEN = 10**4


# filters
CUT_EMPTY_ELEMS = {'article', 'b', 'blockquote', 'dd', 'div', 'dt', 'em',
                   'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'i', 'li', 'main',
                   'p', 'pre', 'q', 'section', 'span', 'strong'}
                   # 'meta', 'td', 'a', 'caption', 'dl', 'header',
                   # 'colgroup', 'col',
#CUT_EMPTY_ELEMS = {'div', 'span'}

# order could matter, using lists to keep extraction deterministic
MANUALLY_CLEANED = [
    # important
    'aside', 'embed', 'footer', 'form', 'head', 'iframe', 'menu', 'object', 'script',
    # other content
    'applet', 'audio', 'canvas', 'figure', 'map', 'picture', 'svg', 'video',
    # secondary
    'area', 'blink', 'button', 'datalist', 'dialog',
    'frame', 'frameset', 'fieldset', 'link', 'input', 'ins', 'label', 'legend',
    'marquee', 'math', 'menuitem', 'nav', 'noscript', 'optgroup', 'option',
    'output', 'param', 'progress', 'rp', 'rt', 'rtc', 'select', 'source',
    'style', 'track', 'textarea', 'time', 'use',
]
# 'meta', 'hr', 'img', 'data', 'details', 'summary'

MANUALLY_STRIPPED = [
    'abbr', 'acronym', 'address', 'bdi', 'bdo', 'big', 'cite', 'data', 'dfn',
    'font', 'hgroup', 'img', 'ins', 'mark', 'meta', 'ruby', 'small', 'tbody',
    'template', 'tfoot', 'thead',
]
# 'center', 'rb', 'wbr'

BASIC_CLEAN_XPATH = XPath(".//aside|.//footer|.//script|.//style")

TAG_CATALOG = frozenset(['blockquote', 'code', 'del', 'head', 'hi', 'lb', 'list', 'p', 'pre', 'quote'])
# + list(CUT_EMPTY_ELEMS)


JUSTEXT_LANGUAGES = {
    'ar': 'Arabic',
    'bg': 'Bulgarian',
    'cz': 'Czech',
    'da': 'Danish',
    'de': 'German',
    'en': 'English',
    'el': 'Greek',
    'es': 'Spanish',
    'fa': 'Persian',
    'fi': 'Finnish',
    'fr': 'French',
    'hr': 'Croatian',
    'hu': 'Hungarian',
    # 'ja': '',
    'ko': 'Korean',
    'id': 'Indonesian',
    'it': 'Italian',
    'no': 'Norwegian_Nynorsk',
    'nl': 'Dutch',
    'pl': 'Polish',
    'pt': 'Portuguese',
    'ro': 'Romanian',
    'ru': 'Russian',
    'sk': 'Slovak',
    'sl': 'Slovenian',
    'sr': 'Serbian',
    'sv': 'Swedish',
    'tr': 'Turkish',
    'uk': 'Ukrainian',
    'ur': 'Urdu',
    'vi': 'Vietnamese',
    # 'zh': '',
}



"""Data structures for handling HTML Elements."""
from typing import Tuple

from haruka_parser.table_processing.html_properties import (
    Display, HorizontalAlignment, VerticalAlignment, WhiteSpace)


class HtmlElement:
    """The HtmlElement class stores properties and metadata of HTML elements.

    Attributes:
    - canvas: the canvas to which the HtmlElement writes its content.
    - tag: tag name of the given HtmlElement.
    - prefix: specifies a prefix that to insert before the tag's content.
    - suffix: a suffix to append after the tag's content.
    - display: :class:`html_properties.Display` strategy used for
      the content.
    - margin_before: vertical margin before the tag's content.
    - margin_after: vertical margin after the tag's content.
    - padding_inline: horizontal padding_inline before the tag's content.
    - whitespace: the :class:`html_properties.Whitespace` handling
      strategy.
    - limit_whitespace_affixes: limit printing of whitespace affixes to
      elements with `normal` whitespace handling.
    - align: the element's horizontal alignment.
    - valign: the element's vertical alignment.
    - previous_margin_after: the margin after of the previous HtmlElement.
    - annotation: annotations associated with the HtmlElement.
    """

    __slots__ = (
        "canvas",
        "tag",
        "prefix",
        "suffix",
        "display",
        "margin_before",
        "margin_after",
        "padding_inline",
        "list_bullet",
        "whitespace",
        "limit_whitespace_affixes",
        "align",
        "valign",
        "previous_margin_after",
        "annotation",
    )

    def __init__(
        self,
        tag: str = "default",
        prefix: str = "",
        suffix: str = "",
        display: Display = Display.inline,
        margin_before: int = 0,
        margin_after: int = 0,
        padding_inline: int = 0,
        list_bullet: str = "",
        whitespace: WhiteSpace = None,
        limit_whitespace_affixes: bool = False,
        align: HorizontalAlignment = HorizontalAlignment.left,
        valign: VerticalAlignment = VerticalAlignment.middle,
        annotation: Tuple[str] = (),
    ):
        self.canvas = None
        self.tag = tag
        self.prefix = prefix
        self.suffix = suffix
        self.display = display
        self.margin_before = margin_before
        self.margin_after = margin_after
        self.padding_inline = padding_inline
        self.list_bullet = list_bullet
        self.whitespace = whitespace
        self.limit_whitespace_affixes = limit_whitespace_affixes
        self.align = align
        self.valign = valign
        self.previous_margin_after = 0
        self.annotation = annotation

    def __copy__(self) -> "HtmlElement":
        """Performance-optimized copy implementation."""
        copy = self.__class__.__new__(self.__class__)
        for attr in self.__slots__:
            setattr(copy, attr, getattr(self, attr))
        return copy

    def write(self, text: str):
        """Write the given HTML text to the element's canvas."""
        if not text or self.display == Display.none:
            return
        self.canvas.write(self, "".join((self.prefix, text, self.suffix)))

    def set_canvas(self, canvas) -> "HtmlElement":
        self.canvas = canvas
        return self

    def set_tag(self, tag: str) -> "HtmlElement":
        self.tag = tag
        return self

    def write_verbatim_text(self, text: str):
        """Write the given text with `Whitespace.pre` to the canvas.

        Args:
            text: the text to write
        """
        if not text:
            return

        if self.display == Display.block:
            self.canvas.open_block(self)

        self.canvas.write(self, text, whitespace=WhiteSpace.pre)

        if self.display == Display.block:
            self.canvas.close_block(self)

    def get_refined_html_element(self, new: "HtmlElement") -> "HtmlElement":
        """Compute the new HTML element based on the previous one.

        Adaptations:
            margin_top: additional margin required when considering
                        margin_bottom of the previous element

        Args:
            new: The new HtmlElement to be applied to the current context.

        Returns:
            The refined element with the context applied.
        """
        new.canvas = self.canvas

        # inherit `display:none` attributes and ignore further refinements
        if self.display == Display.none:
            new.display = Display.none
            return new

        # no whitespace set => inherit
        new.whitespace = new.whitespace or self.whitespace

        # do not display whitespace only affixes in Whitespace.pre areas
        # if `limit_whitespace_affixes` is set.
        if new.limit_whitespace_affixes and self.whitespace == WhiteSpace.pre:
            if new.prefix.isspace():
                new.prefix = ""
            if new.suffix.isspace():
                new.suffix = ""

        if new.display == Display.block and self.display == Display.block:
            new.previous_margin_after = self.margin_after

        return new

    def __str__(self) -> str:
        return (
            f"<{self.tag} prefix={self.prefix}, suffix={self.suffix}, "
            f"display={self.display}, margin_before={self.margin_before}, "
            f"margin_after={self.margin_after}, "
            f"padding_inline={self.padding_inline}, "
            f"list_bullet={self.list_bullet}, "
            f"whitespace={self.whitespace}, align={self.align}, "
            f"valign={self.valign}, annotation={self.annotation}>"
        )

    __repr__ = __str__


"""
An empty default HTML element.
"""
DEFAULT_HTML_ELEMENT = HtmlElement()



#!/usr/bin/env python3
# encoding: utf-8

"""HTML attribute handling."""
from contextlib import suppress
from re import compile as re_compile
from typing import Dict

from haruka_parser.table_processing.html_element import HtmlElement
from haruka_parser.table_processing.html_properties import (
    Display, HorizontalAlignment, VerticalAlignment, WhiteSpace)


class CssParse:
    """Parse CSS specifications and applies them to HtmlElements.

    The attribute `display: none`, for instance, is translated to
    :attr:`HtmlElement.display=Display.none`.
    """

    # used to separate value and unit from each other
    RE_UNIT = re_compile(r"(-?[0-9.]+)(\w+)")

    @staticmethod
    def attr_style(style_attribute: str, html_element: HtmlElement):
        """Apply the provided style attributes to the given HtmlElement.

        Args:
          style_attribute: The attribute value of the given style sheet.
                           Example: display: none
          html_element: The HtmlElement to which the given style is applied.
        """
        for style_directive in style_attribute.lower().split(";"):
            if ":" not in style_directive:
                continue
            key, value = (s.strip() for s in style_directive.split(":", 1))

            try:
                apply_style = getattr(
                    CssParse, "attr_" + key.replace("-webkit-", "").replace("-", "_")
                )
                apply_style(value, html_element)
            except AttributeError:
                pass

    @staticmethod
    def _get_em(length: str) -> int:
        """Convert length specifications into em.

        This function takes a length specification (e.g., 2em, 2px, etc.) and
        transforms it into em.

        Args:
          length: the length specification.

        Returns:
            the length in em.
        """
        _m = CssParse.RE_UNIT.search(length)
        value = float(_m.group(1))
        unit = _m.group(2)

        if unit not in ("em", "qem", "rem"):
            return int(round(value / 8))
        return int(round(value))

    # ------------------------------------------------------------------------
    # css styles
    # ------------------------------------------------------------------------

    @staticmethod
    def attr_display(value: str, html_element: HtmlElement):
        """Apply the given display value."""
        if html_element.display == Display.none:
            return

        if value == "block":
            html_element.display = Display.block
        elif value == "none":
            html_element.display = Display.none
        else:
            html_element.display = Display.inline

    @staticmethod
    def attr_white_space(value: str, html_element: HtmlElement):
        """Apply the given white-space value."""
        if value in ("normal", "nowrap"):
            html_element.whitespace = WhiteSpace.normal
        elif value in ("pre", "pre-line", "pre-wrap"):
            html_element.whitespace = WhiteSpace.pre

    @staticmethod
    def attr_margin_top(value: str, html_element: HtmlElement):
        """Apply the given top margin."""
        with suppress(ValueError):
            html_element.margin_before = CssParse._get_em(value)

    @staticmethod
    def attr_margin_bottom(value: str, html_element: HtmlElement):
        """Apply the provided bottom margin."""
        with suppress(ValueError):
            html_element.margin_after = CssParse._get_em(value)

    @staticmethod
    def attr_padding_left(value: str, html_element: HtmlElement):
        """Apply the given left padding_inline."""
        with suppress(ValueError):
            html_element.padding_inline = CssParse._get_em(value)

    @staticmethod
    def attr_horizontal_align(value: str, html_element: HtmlElement):
        """Apply the provided horizontal alignment."""
        with suppress(KeyError):
            html_element.align = HorizontalAlignment[value]

    @staticmethod
    def attr_vertical_align(value: str, html_element: HtmlElement):
        """Apply the given vertical alignment."""
        with suppress(KeyError):
            html_element.valign = VerticalAlignment[value]

    # register aliases
    attr_margin_before = attr_margin_top
    attr_margin_after = attr_margin_bottom
    attr_padding_start = attr_padding_left


DEFAULT_ATTRIBUTE_MAP = {
    "style": CssParse.attr_style,
    "align": CssParse.attr_horizontal_align,
    "valign": CssParse.attr_vertical_align,
}

class Attribute:
    """Handle HTML attributes such as `align`, and `valign`.

    This class handles HTML attributes by mapping them to the corresponding
    functions in the :class:`CssParse` class.

    Attributes:
        attribute_mapping: a mapping of attributes to the corresponding handler
                           functions.
    """

    def __init__(self):
        self.attribute_mapping = DEFAULT_ATTRIBUTE_MAP

    def apply_attributes(
        self, attributes: Dict[str, str], html_element: HtmlElement
    ) -> HtmlElement:
        """Apply the attributes to the given HTML element.

        Args:
            attributes: the list of attributes
            html_element: the HTML element for which the attributes are parsed
        """
        supported_attributes = (
            (name, val)
            for name, val in attributes.items()
            if name in self.attribute_mapping
        )
        for attr_name, attr_value in supported_attributes:
            self.attribute_mapping[attr_name](attr_value, html_element)
        return html_element

    def merge_attribute_map(self) -> None:
        
        pass


import ctypes
import json
import os
from platform import machine
from sys import platform

if platform == 'darwin':
    file_ext = '-arm64.dylib' if machine() == "arm64" else '-x86.dylib'
elif platform in ('win32', 'cygwin'):
    file_ext = '-win64.dll' if 8 == ctypes.sizeof(ctypes.c_voidp) else '-win32.dll'
else:
    if machine() == "aarch64":
        file_ext = '-arm64.so'
    elif "x86" in machine():
        file_ext = '-x86.so'
    else:
        file_ext = '-amd64.so'


root_dir = os.path.abspath(os.path.dirname(__file__))
library = ctypes.cdll.LoadLibrary(f'{root_dir}/lib/libparser{file_ext}')

parse = library.parse
parse.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
parse.restype = ctypes.c_char_p

freeMemory = library.freeMemory
freeMemory.argtypes = [ctypes.c_char_p]

def readablity_go_parse(html_content, url="https://example.com"):
    result = parse(html_content.encode('utf-8'), url.encode('utf-8'))
    data = json.loads(ctypes.c_char_p(result).value.decode('utf-8'))
    freeMemory(data["id"].encode('utf-8'))
    return data


"""
Functions dedicated to command-line processing.
"""

import gzip
import logging
import random
import re
import string
import sys
import traceback

from base64 import urlsafe_b64encode
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from datetime import datetime
from functools import partial
from os import makedirs, path, stat, walk

from courlan import UrlStore, extract_domain, get_base_url  # validate_url

from trafilatura import spider

from .baseline import html2txt
from .core import extract
from .deduplication import generate_bow_hash
from .downloads import (add_to_compressed_dict, buffered_downloads,
                        load_download_buffer)
from .feeds import find_feed_urls
from .meta import reset_caches
from .settings import FILENAME_LEN, MAX_FILES_PER_DIRECTORY, args_to_extractor
from .sitemaps import sitemap_search
from .utils import LANGID_FLAG, URL_BLACKLIST_REGEX, is_acceptable_length, language_classifier, make_chunks


LOGGER = logging.getLogger(__name__)

random.seed(345)  # make generated file names reproducible
CHAR_CLASS = string.ascii_letters + string.digits

STRIP_DIR = re.compile(r'[^/]+$')
STRIP_EXTENSION = re.compile(r'\.[a-z]{2,5}$')

CLEAN_XML = re.compile(r"<[^<]+?>")

INPUT_URLS_ARGS = ['URL', 'crawl', 'explore', 'probe', 'feed', 'sitemap']

EXTENSION_MAPPING = {
    'csv': '.csv',
    'json': '.json',
    'xml': '.xml',
    'xmltei': '.xml',
}


def load_input_urls(args):
    '''Read list of URLs to process or derive one from command-line arguments'''
    input_urls = []

    if args.input_file:
        try:
            # optional: errors='strict', buffering=1
            with open(args.input_file, mode='r', encoding='utf-8') as inputfile:
                input_urls.extend(line.strip() for line in inputfile)
        except UnicodeDecodeError:
            sys.exit('ERROR: system, file type or buffer encoding')
    else:
        for arg in INPUT_URLS_ARGS:
            if getattr(args, arg):
                input_urls = [getattr(args, arg)]
                break

    if not input_urls:
        LOGGER.warning('No input provided')

    # uniq URLs while preserving order (important)
    return list(dict.fromkeys(input_urls))


def load_blacklist(filename):
    '''Read list of unwanted URLs'''
    with open(filename, 'r', encoding='utf-8') as inputfh:
        # if validate_url(url)[0] is True:
        blacklist = {URL_BLACKLIST_REGEX.sub('', line.strip()) for line in inputfh}
    return blacklist


def load_input_dict(args):
    '''Read input list of URLs to process and
       build a domain-aware dictionary'''
    inputlist = load_input_urls(args)
    # deduplicate, filter and convert to dict
    return add_to_compressed_dict(
        inputlist,
        blacklist=args.blacklist,
        compression=(args.sitemap and not args.list),
        url_filter=args.url_filter,
        verbose=args.verbose
    )


def check_outputdir_status(directory):
    '''Check if the output directory is within reach and writable'''
    # check the directory status
    if not path.exists(directory) or not path.isdir(directory):
        try:
            makedirs(directory, exist_ok=True)
        except OSError:
            # maybe the directory has already been created
            #sleep(0.25)
            #if not path.exists(directory) or not path.isdir(directory):
            sys.stderr.write('ERROR: Destination directory cannot be created: ' + directory + '\n')
            # raise OSError()
            return False
    return True


def determine_counter_dir(dirname, counter):
    '''Return a destination directory based on a file counter'''
    if counter is not None:
        counter_dir = str(int(counter/MAX_FILES_PER_DIRECTORY) + 1)
    else:
        counter_dir = ''
    return path.join(dirname, counter_dir)


def get_writable_path(destdir, extension):
    '''Find a writable path and return it along with its random file name'''
    output_path = None
    while output_path is None or path.exists(output_path):
        # generate a random filename of the desired length
        filename = ''.join(random.choice(CHAR_CLASS) for _ in range(FILENAME_LEN))
        output_path = path.join(destdir, filename + extension)
    return output_path, filename


def generate_hash_filename(content: str) -> str:
    """Create a filename-safe string by hashing the given content
    after deleting potential XML tags."""
    return urlsafe_b64encode(
               generate_bow_hash(CLEAN_XML.sub("", content), 12)
           ).decode()


def determine_output_path(args, orig_filename, content, counter=None, new_filename=None):
    '''Pick a directory based on selected options and a file name based on output type'''
    # determine extension, TXT by default
    extension = EXTENSION_MAPPING.get(args.output_format, '.txt')

    if args.keep_dirs:
        # strip directory
        original_dir = STRIP_DIR.sub('', orig_filename)
        destination_dir = path.join(args.output_dir, original_dir)
        # strip extension
        filename = STRIP_EXTENSION.sub('', orig_filename)
    else:
        destination_dir = determine_counter_dir(args.output_dir, counter)
        # use cryptographic hash on file contents to define name
        filename = new_filename or generate_hash_filename(content)

    output_path = path.join(destination_dir, filename + extension)
    return output_path, destination_dir


def archive_html(htmlstring, args, counter=None):
    '''Write a copy of raw HTML in backup directory'''
    destination_directory = determine_counter_dir(args.backup_dir, counter)
    output_path, filename = get_writable_path(destination_directory, '.html.gz')
    # check the directory status
    if check_outputdir_status(destination_directory) is True:
        # write
        with gzip.open(output_path, 'wb') as outputfile:
            outputfile.write(htmlstring.encode('utf-8'))
    return filename


def write_result(result, args, orig_filename=None, counter=None, new_filename=None):
    '''Deal with result (write to STDOUT or to file)'''
    if result is None:
        return
    if args.output_dir is None:
        sys.stdout.write(result + '\n')
    else:
        destination_path, destination_dir = determine_output_path(args, orig_filename, result, counter, new_filename)
        # check the directory status
        if check_outputdir_status(destination_dir) is True:
            with open(destination_path, mode='w', encoding='utf-8') as outputfile:
                outputfile.write(result)


def generate_filelist(inputdir):
    '''Walk the directory tree and output all file names'''
    for root, _, inputfiles in walk(inputdir):
        for fname in inputfiles:
            yield path.join(root, fname)


def file_processing(filename, args, counter=None, options=None):
    '''Aggregated functions to process a file in a list'''
    if not options:
        options = args_to_extractor(args)
    options.source = filename

    with open(filename, 'rb') as inputf:
        htmlstring = inputf.read()

    file_stat = stat(filename)
    ref_timestamp = min(file_stat.st_ctime, file_stat.st_mtime)
    options.date_params["max_date"] = datetime.fromtimestamp(ref_timestamp).strftime("%Y-%m-%d")

    result = examine(htmlstring, args, options=options)
    write_result(result, args, filename, counter, new_filename=None)


def process_result(htmlstring, args, counter, options):
    '''Extract text and metadata from a download webpage and eventually write out the result'''
    # backup option
    fileslug = archive_html(htmlstring, args, counter) if args.backup_dir else None
    # process
    result = examine(htmlstring, args, options=options)
    write_result(result, args, orig_filename=fileslug, counter=counter, new_filename=fileslug)
    # increment written file counter
    if counter is not None and result is not None:
        counter += 1
    return counter


def download_queue_processing(url_store, args, counter, options):
    '''Implement a download queue consumer, single- or multi-threaded'''
    errors = []
    while url_store.done is False:
        bufferlist, url_store = load_download_buffer(url_store, options.config.getfloat('DEFAULT', 'SLEEP_TIME'))
        # process downloads
        for url, result in buffered_downloads(bufferlist, args.parallel, options=options):
            # handle result
            if result is not None:
                options.url = url
                counter = process_result(result, args, counter, options)
            else:
                LOGGER.warning('No result for URL: %s', url)
                errors.append(url)
    return errors, counter


def cli_discovery(args):
    "Group CLI functions dedicated to URL discovery."
    url_store = load_input_dict(args)
    input_urls = url_store.dump_urls()
    if args.list:
        url_store.reset()

    options = args_to_extractor(args)
    func = partial(
               find_feed_urls if args.feed else sitemap_search,
               target_lang=args.target_language,
               external=options.config.getboolean('DEFAULT', 'EXTERNAL_URLS'),
               sleep_time=options.config.getfloat('DEFAULT', 'SLEEP_TIME')
           )

    # link discovery and storage
    with ThreadPoolExecutor(max_workers=args.parallel) as executor:
        futures = (executor.submit(func, url) for url in input_urls)
        # process results from the parallel threads and add them
        # to the compressed URL dictionary for further processing
        for future in as_completed(futures):
            if future.result() is not None:
                url_store.add_urls(future.result())
                # empty buffer in order to spare memory
                if args.sitemap and args.list and len(url_store.get_known_domains()) >= args.parallel:
                    url_store.print_unvisited_urls()
                    url_store.reset()
                    reset_caches()

    # process the (rest of the) links found
    error_caught = url_processing_pipeline(args, url_store)

    # activate site explorer
    if args.explore:
        # add to compressed dict and crawl the remaining websites
        control_dict = build_exploration_dict(url_store, input_urls, args)
        cli_crawler(args, url_store=control_dict, options=options)


def build_exploration_dict(url_store, input_urls, args):
    "Find domains for which nothing has been found and add info to the crawl dict."
    input_domains = {extract_domain(u) for u in input_urls}
    known_domains = {extract_domain(u) for u in url_store.get_known_domains()}
    still_to_crawl = input_domains - known_domains
    new_input_urls = [u for u in input_urls if extract_domain(u) in still_to_crawl]
    control_dict = add_to_compressed_dict(
                       new_input_urls,
                       blacklist=args.blacklist,
                       url_filter=args.url_filter,
                       verbose=args.verbose
                   )
    return control_dict


def cli_crawler(args, n=30, url_store=None, options=None):
    '''Start a focused crawler which downloads a fixed number of URLs within a website
       and prints the links found in the process'''
    if not options:
        options = args_to_extractor(args)
    sleep_time = options.config.getfloat('DEFAULT', 'SLEEP_TIME')
    # counter = None
    # load input URLs
    if url_store is None:
        spider.URL_STORE.add_urls(load_input_urls(args))
    else:
        spider.URL_STORE = url_store
    # load crawl data
    for hostname in spider.URL_STORE.get_known_domains():
        if spider.URL_STORE.urldict[hostname].tuples:
            startpage = spider.URL_STORE.get_url(hostname, as_visited=False)
            # base_url, i, known_num, rules, is_on
            _ = spider.init_crawl(startpage, None, set(), language=args.target_language)
            # update info
            # TODO: register changes?
            # if base_url != hostname:
            # ...
    # iterate until the threshold is reached
    while spider.URL_STORE.done is False:
        bufferlist, spider.URL_STORE = load_download_buffer(spider.URL_STORE, sleep_time)
        # start several threads
        for url, result in buffered_downloads(bufferlist, args.parallel, decode=False, options=options):
            base_url = get_base_url(url)
            # handle result
            if result is not None:
                spider.process_response(result, base_url, args.target_language, rules=spider.URL_STORE.get_rules(base_url))
                # just in case a crawl delay is specified in robots.txt
                # sleep(spider.get_crawl_delay(spider.URL_STORE.get_rules(base_url)))
        # early exit if maximum count is reached
        if any(c >= n for c in spider.URL_STORE.get_all_counts()):
            break
    # print results
    print('\n'.join(u for u in spider.URL_STORE.dump_urls()))
    #return todo, known_links


def probe_homepage(args):
    "Probe websites for extractable content and print the fitting ones."
    input_urls = load_input_urls(args)
    options = args_to_extractor(args)

    for url, result in buffered_downloads(input_urls, args.parallel, options=options):
        if result is not None:
            result = html2txt(result)
            if result and len(result) > options.min_extracted_size and any(c.isalpha() for c in result):
                if not LANGID_FLAG or not args.target_language or language_classifier(result, "") == args.target_language:
                    print(url, flush=True)


def url_processing_pipeline(args, url_store):
    '''Aggregated functions to show a list and download and process an input list'''
    # print list without further processing
    if args.list:
        url_store.print_unvisited_urls()  # and not write_result()
        return False  # and not sys.exit(0)

    options = args_to_extractor(args)

    # initialize file counter if necessary
    if url_store.total_url_number() > MAX_FILES_PER_DIRECTORY:
        counter = 0
    else:
        counter = None
    # download strategy
    errors, counter = download_queue_processing(url_store, args, counter, options)
    LOGGER.debug('%s URLs could not be found', len(errors))
    # option to retry
    if args.archived is True:
        url_store = UrlStore()
        url_store.add_urls(['https://web.archive.org/web/20/' + e for e in errors])
        if len(url_store.find_known_urls('https://web.archive.org')) > 0:
            archived_errors, _ = download_queue_processing(url_store, args, counter, options)
            LOGGER.debug('%s archived URLs out of %s could not be found', len(archived_errors), len(errors))
            # pass information along if URLs are missing
            return bool(archived_errors)
    # pass information along if URLs are missing
    return bool(errors)


def file_processing_pipeline(args):
    '''Define batches for parallel file processing and perform the extraction'''
    filecounter = None
    options = args_to_extractor(args)
    timeout = options.config.getint('DEFAULT', 'EXTRACTION_TIMEOUT')

    # max_tasks_per_child available in Python >= 3.11
    with ProcessPoolExecutor(max_workers=args.parallel) as executor:
        # chunk input: https://github.com/python/cpython/issues/74028
        for filebatch in make_chunks(generate_filelist(args.input_dir), MAX_FILES_PER_DIRECTORY):
            if filecounter is None and len(filebatch) >= MAX_FILES_PER_DIRECTORY:
                filecounter = 0
            worker = partial(file_processing, args=args, counter=filecounter, options=options)
            executor.map(worker, filebatch, chunksize=10, timeout=timeout)
            # update counter
            if filecounter is not None:
                filecounter += len(filebatch)


def examine(htmlstring, args, url=None, options=None):
    """Generic safeguards and triggers"""
    result = None
    if not options:
        options = args_to_extractor(args, url)
    # safety check
    if htmlstring is None:
        sys.stderr.write('ERROR: empty document\n')
    elif not is_acceptable_length(len(htmlstring), options):
        sys.stderr.write('ERROR: file size\n')
    # proceed
    else:
        try:
            result = extract(htmlstring, options=options)
        # ugly but efficient
        except Exception as err:
            sys.stderr.write(f'ERROR: {str(err)}' + '\n' + traceback.format_exc() + '\n')
    return result



"""A list of all supported languages in highlight.js.

This file is automatically generated by tools.
"""
import typing


class Language(typing.NamedTuple):
    """A language supported by highlight.js."""
    name: str
    alias: str


languages: typing.List[Language] = [
    Language(name='1C:Enterprise', alias='1c'),
    Language(name='Augmented Backus-Naur Form', alias='abnf'),
    Language(name='Apache Access Log', alias='accesslog'),
    Language(name='ActionScript', alias='actionscript'),
    Language(name='Ada', alias='ada'),
    Language(name='AngelScript', alias='angelscript'),
    Language(name='Apache config', alias='apache'),
    Language(name='AppleScript', alias='applescript'),
    Language(name='ArcGIS Arcade', alias='arcade'),
    Language(name='Arduino', alias='arduino'),
    Language(name='ARM Assembly', alias='armasm'),
    Language(name='AsciiDoc', alias='asciidoc'),
    Language(name='AspectJ', alias='aspectj'),
    Language(name='AutoHotkey', alias='autohotkey'),
    Language(name='AutoIt', alias='autoit'),
    Language(name='AVR Assembly', alias='avrasm'),
    Language(name='Awk', alias='awk'),
    Language(name='X++', alias='axapta'),
    Language(name='Bash', alias='bash'),
    Language(name='BASIC', alias='basic'),
    Language(name='Backus–Naur Form', alias='bnf'),
    Language(name='Brainfuck', alias='brainfuck'),
    Language(name='C', alias='c'),
    Language(name='C/AL', alias='cal'),
    Language(name='Cap’n Proto', alias='capnproto'),
    Language(name='Ceylon', alias='ceylon'),
    Language(name='Clean', alias='clean'),
    Language(name='Clojure', alias='clojure'),
    Language(name='Clojure REPL', alias='clojure-repl'),
    Language(name='CMake', alias='cmake'),
    Language(name='CoffeeScript', alias='coffeescript'),
    Language(name='Coq', alias='coq'),
    Language(name='Caché Object Script', alias='cos'),
    Language(name='C++', alias='cpp'),
    Language(name='crmsh', alias='crmsh'),
    Language(name='Crystal', alias='crystal'),
    Language(name='C#', alias='csharp'),
    Language(name='CSP', alias='csp'),
    Language(name='CSS', alias='css'),
    Language(name='D', alias='d'),
    Language(name='Dart', alias='dart'),
    Language(name='Delphi', alias='delphi'),
    Language(name='Diff', alias='diff'),
    Language(name='Django', alias='django'),
    Language(name='DNS Zone', alias='dns'),
    Language(name='Dockerfile', alias='dockerfile'),
    Language(name='Batch file (DOS)', alias='dos'),
    Language(name='dsconfig', alias='dsconfig'),
    Language(name='Device Tree', alias='dts'),
    Language(name='Dust', alias='dust'),
    Language(name='Extended Backus-Naur Form', alias='ebnf'),
    Language(name='Elixir', alias='elixir'),
    Language(name='Elm', alias='elm'),
    Language(name='ERB', alias='erb'),
    Language(name='Erlang', alias='erlang'),
    Language(name='Erlang REPL', alias='erlang-repl'),
    Language(name='Excel formulae', alias='excel'),
    Language(name='FIX', alias='fix'),
    Language(name='Flix', alias='flix'),
    Language(name='Fortran', alias='fortran'),
    Language(name='F#', alias='fsharp'),
    Language(name='GAMS', alias='gams'),
    Language(name='GAUSS', alias='gauss'),
    Language(name='G-code (ISO 6983)', alias='gcode'),
    Language(name='Gherkin', alias='gherkin'),
    Language(name='GLSL', alias='glsl'),
    Language(name='GML', alias='gml'),
    Language(name='Go', alias='go'),
    Language(name='Golo', alias='golo'),
    Language(name='Gradle', alias='gradle'),
    Language(name='GraphQL', alias='graphql'),
    Language(name='Groovy', alias='groovy'),
    Language(name='HAML', alias='haml'),
    Language(name='Handlebars', alias='handlebars'),
    Language(name='Haskell', alias='haskell'),
    Language(name='Haxe', alias='haxe'),
    Language(name='HSP', alias='hsp'),
    Language(name='HTTP', alias='http'),
    Language(name='Hy', alias='hy'),
    Language(name='Inform 7', alias='inform7'),
    Language(name='TOML, also INI', alias='ini'),
    Language(name='IRPF90', alias='irpf90'),
    Language(name='ISBL', alias='isbl'),
    Language(name='Java', alias='java'),
    Language(name='JavaScript', alias='javascript'),
    Language(name='JBoss CLI', alias='jboss-cli'),
    Language(name='JSON', alias='json'),
    Language(name='Julia', alias='julia'),
    Language(name='Julia REPL', alias='julia-repl'),
    Language(name='Kotlin', alias='kotlin'),
    Language(name='Lasso', alias='lasso'),
    Language(name='LaTeX', alias='latex'),
    Language(name='LDIF', alias='ldif'),
    Language(name='Leaf', alias='leaf'),
    Language(name='Less', alias='less'),
    Language(name='Lisp', alias='lisp'),
    Language(name='LiveCode', alias='livecodeserver'),
    Language(name='LiveScript', alias='livescript'),
    Language(name='LLVM IR', alias='llvm'),
    Language(name='LSL (Linden Scripting Language)', alias='lsl'),
    Language(name='Lua', alias='lua'),
    Language(name='Makefile', alias='makefile'),
    Language(name='Markdown', alias='markdown'),
    Language(name='Mathematica', alias='mathematica'),
    Language(name='Matlab', alias='matlab'),
    Language(name='Maxima', alias='maxima'),
    Language(name='MEL', alias='mel'),
    Language(name='Mercury', alias='mercury'),
    Language(name='MIPS Assembly', alias='mipsasm'),
    Language(name='Mizar', alias='mizar'),
    Language(name='Mojolicious', alias='mojolicious'),
    Language(name='Monkey', alias='monkey'),
    Language(name='MoonScript', alias='moonscript'),
    Language(name='N1QL', alias='n1ql'),
    Language(name='Nested Text', alias='nestedtext'),
    Language(name='Nginx config', alias='nginx'),
    Language(name='Nim', alias='nim'),
    Language(name='Nix', alias='nix'),
    Language(name='Node REPL', alias='node-repl'),
    Language(name='NSIS', alias='nsis'),
    Language(name='Objective-C', alias='objectivec'),
    Language(name='OCaml', alias='ocaml'),
    Language(name='OpenSCAD', alias='openscad'),
    Language(name='Oxygene', alias='oxygene'),
    Language(name='Parser3', alias='parser3'),
    Language(name='Perl', alias='perl'),
    Language(name='Packet Filter config', alias='pf'),
    Language(name='PostgreSQL', alias='pgsql'),
    Language(name='php', alias='php'),
    Language(name='PHP template', alias='php-template'),
    Language(name='Plain text', alias='plaintext'),
    Language(name='Pony', alias='pony'),
    Language(name='PowerShell', alias='powershell'),
    Language(name='Processing', alias='processing'),
    Language(name='Python profiler', alias='profile'),
    Language(name='Prolog', alias='prolog'),
    Language(name='.properties', alias='properties'),
    Language(name='Protocol Buffers', alias='protobuf'),
    Language(name='Puppet', alias='puppet'),
    Language(name='PureBASIC', alias='purebasic'),
    Language(name='Python', alias='python'),
    Language(name='python-repl', alias='python-repl'),
    Language(name='Q', alias='q'),
    Language(name='QML', alias='qml'),
    Language(name='R', alias='r'),
    Language(name='ReasonML', alias='reasonml'),
    Language(name='RenderMan RIB', alias='rib'),
    Language(name='Roboconf', alias='roboconf'),
    Language(name='MikroTik RouterOS script', alias='routeros'),
    Language(name='RenderMan RSL', alias='rsl'),
    Language(name='Ruby', alias='ruby'),
    Language(name='Oracle Rules Language', alias='ruleslanguage'),
    Language(name='Rust', alias='rust'),
    Language(name='SAS', alias='sas'),
    Language(name='Scala', alias='scala'),
    Language(name='Scheme', alias='scheme'),
    Language(name='Scilab', alias='scilab'),
    Language(name='SCSS', alias='scss'),
    Language(name='Shell Session', alias='shell'),
    Language(name='Smali', alias='smali'),
    Language(name='Smalltalk', alias='smalltalk'),
    Language(name='SML (Standard ML)', alias='sml'),
    Language(name='SQF', alias='sqf'),
    Language(name='SQL', alias='sql'),
    Language(name='Stan', alias='stan'),
    Language(name='Stata', alias='stata'),
    Language(name='STEP Part 21', alias='step21'),
    Language(name='Stylus', alias='stylus'),
    Language(name='SubUnit', alias='subunit'),
    Language(name='Swift', alias='swift'),
    Language(name='Tagger Script', alias='taggerscript'),
    Language(name='Test Anything Protocol', alias='tap'),
    Language(name='Tcl', alias='tcl'),
    Language(name='Thrift', alias='thrift'),
    Language(name='TP', alias='tp'),
    Language(name='Twig', alias='twig'),
    Language(name='TypeScript', alias='typescript'),
    Language(name='Vala', alias='vala'),
    Language(name='Visual Basic .NET', alias='vbnet'),
    Language(name='VBScript', alias='vbscript'),
    Language(name='VBScript in HTML', alias='vbscript-html'),
    Language(name='Verilog', alias='verilog'),
    Language(name='VHDL', alias='vhdl'),
    Language(name='Vim Script', alias='vim'),
    Language(name='WebAssembly', alias='wasm'),
    Language(name='Wren', alias='wren'),
    Language(name='Intel x86 Assembly', alias='x86asm'),
    Language(name='XL', alias='xl'),
    Language(name='HTML, XML', alias='xml'),
    Language(name='XQuery', alias='xquery'),
    Language(name='YAML', alias='yaml'),
    Language(name='Zephir', alias='zephir'),
]


# -*- coding: utf-8 -*-
"""
Python & command-line tool to gather text on the Web:
web crawling/scraping, extraction of text, metadata, comments.
"""


__title__ = 'trafilatura'
__author__ = 'Adrien Barbaresi and contributors'
__license__ = "Apache-2.0"
__copyright__ = 'Copyright 2019-2024, Adrien Barbaresi'
__version__ = '1.10.0'


import logging

from .baseline import baseline, html2txt
from .core import bare_extraction, extract, process_record
from .downloads import fetch_response, fetch_url
from .metadata import extract_metadata
from .utils import load_html

logging.getLogger(__name__).addHandler(logging.NullHandler())



from typing import Callable, Dict

from resiliparse.parse.html import HTMLTree

from haruka_parser.line_processing import have_chinese_characters
from haruka_parser.table_processing.canvas import Canvas
from haruka_parser.table_processing.html_state import (HtmlDocumentState,
                                                       ParserConfig)
from haruka_parser.table_processing.tags.table_tag import (table_end_handler,
                                                           table_start_handler,
                                                           td_end_handler,
                                                           td_start_handler,
                                                           tr_start_handler)
from haruka_parser.line_processing import restore_replacements

def remove_consecutive_blank_lines(table_text: str) -> str:

    lines = [line for line in table_text.split("\n") if line.strip()]

    if not lines:
        return ""

    space_map = [[1 if char == " " else 0 for char in line] for line in lines]

    max_len = max([len(line) for line in lines])

    normalized_space_map = [row + [0] * (max_len - len(row)) for row in space_map]

    column_space_count = [0] * max_len
    for col in range(max_len):
        for row in range(len(normalized_space_map)):
            if normalized_space_map[row][col] == 1:
                column_space_count[col] += 1

    column_all_space = [col for col in range(max_len) if column_space_count[col] == len(normalized_space_map)]

    column_to_append = {}
    column_to_del = []

    # Reduce consecutive columns with spaces to a maximum of 3
    MAX_SPACE_COLUMN = 3
    i = 0
    while i < len(column_all_space):
        start = i
        while (
            i < len(column_all_space) - 1
            and column_all_space[i] + 1 == column_all_space[i + 1]
        ):
            i += 1
        end = i

        if column_all_space[start] == 0:
            for j in range(start, end + 1):
                column_to_del.append(column_all_space[j])
        elif end - start + 1 > MAX_SPACE_COLUMN:
            for j in range(start + MAX_SPACE_COLUMN, end + 1):
                column_to_del.append(column_all_space[j])
        elif end - start + 1 < MAX_SPACE_COLUMN:
            column_to_append[column_all_space[start]] = MAX_SPACE_COLUMN - (end - start + 1)
        i += 1

    optimized_lines = []
    for line in lines:
        optimized_line = ""
        for i, char in enumerate(line):
            if i in column_to_del:
                continue
            # currently not append space
            # if i in column_to_append:
            #     optimized_line += " " * column_to_append[i]
            optimized_line += char
        optimized_lines.append(optimized_line)

    optimized_table_text = "\n".join(optimized_lines)

    return optimized_table_text

class TableExtractor():

    def __init__(self, html_tree, config: ParserConfig = None) -> None:

        config = config or ParserConfig()

        # setup start and end tag call tables
        self.start_tag_handler_dict: Dict[
            str, Callable[[HtmlDocumentState, Dict], None]
        ] = {
            "table": table_start_handler,
            "tr": tr_start_handler,
            "td": td_start_handler,
            "th": td_start_handler,
            # "ul": ul_start_handler,
            # "ol": ol_start_handler,
            # "li": li_start_handler,
            # "br": br_start_handler,
            # "a": a_start_handler if config.parse_a() else None,
            # "img": img_start_handler if config.display_images else None,
        }
        self.end_tag_handler_dict: Dict[str, Callable[[HtmlDocumentState], None]] = {
            "table": table_end_handler,
            # "ul": ul_end_handler,
            # "ol": ol_end_handler,
            "td": td_end_handler,
            "th": td_end_handler,
            # "a": a_end_handler if config.parse_a() else None,
        }

        self.canvas = self._parse_html_tree(HtmlDocumentState(config), html_tree)

    def _parse_html_tree(self, state: HtmlDocumentState, tree) -> Canvas:
        """Parse the HTML tree.

        Args:
            tree: the HTML tree to parse.
        """
        if isinstance(tree.tag, str):
            attrs = {}
            if tree.attrs:
                for attr in tree.attrs:
                    attrs[attr] = tree.getattr(attr)
            state.apply_starttag_layout(tree.tag, attrs)

            if handler := self.start_tag_handler_dict.get(tree.tag):
                handler(state, attrs)
            cur = state.tags[-1]
            cur.canvas.open_tag(cur)

            state.tags[-1].write(tree.value)

            for node in tree.child_nodes:
                self._parse_html_tree(state, node)

            # handle the endtag
            if handler := self.end_tag_handler_dict.get(tree.tag):
                handler(state)
            prev = state.tags.pop()
            prev.canvas.close_tag(prev)

        return state.canvas

    def get_text(self) -> str:
        """Return the text extracted from the HTML page."""
        table_text = self.canvas.get_text()
        try:
            table_text = remove_consecutive_blank_lines(table_text)
        except Exception as e:
            import traceback
            traceback.print_exc()
        return table_text.replace(" ", "[HARUKA_PARSER_BS]").replace(
            "\n", "[HARUKA_PARSER_CHANGE_LINE]"
        )


# Extracting text including tables from HTML
def extract_tables(tree, replacement_manager, config, info):

    # need to think about nested table?
    tables = tree.document.query_selector_all("table")

    for table in tables:
        if table.parent:
            info['table'] += 1
            table_tree = HTMLTree.parse(restore_replacements(str(table), replacement_manager, config, info))
            table_text = TableExtractor(table_tree.body).get_text()
            if have_chinese_characters(table_text):
                info["chinese_table"] += 1
            new_p = tree.create_element('pre')
            new_p.html = replacement_manager.add_replacement(
                table_text, tag="table"
            )
            table.parent.replace_child(new_p, table)



import re
from functools import lru_cache
from html import unescape

import html2text
import numpy as np
import yaml

LINES_TRIMMING = re.compile(r'(?<![p{P}>])\n', flags=re.UNICODE|re.MULTILINE)

AUTHOR_PREFIX = re.compile(r'^([a-zäöüß]+(ed|t))? ?(written by|words by|words|by|von|from) ', flags=re.IGNORECASE)
AUTHOR_REMOVE_NUMBERS = re.compile(r'\d.+?$')
AUTHOR_TWITTER = re.compile(r'@[\w]+')
AUTHOR_REPLACE_JOIN = re.compile(r'[._+]')
AUTHOR_REMOVE_NICKNAME = re.compile(r'["‘({\[’\'][^"]+?[‘’"\')\]}]')
AUTHOR_REMOVE_SPECIAL = re.compile(r'[^\w]+$|[:()?*$#!%/<>{}~¿]')
AUTHOR_REMOVE_PREPOSITION = re.compile(r'\b\s+(am|on|for|at|in|to|from|of|via|with|—|-|–)\s+(.*)', flags=re.IGNORECASE)
AUTHOR_EMAIL = re.compile(r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b')
AUTHOR_SPLIT = re.compile(r'/|;|,|\||&|(?:^|\W)[u|a]nd(?:$|\W)', flags=re.IGNORECASE)
AUTHOR_EMOJI_REMOVE = re.compile(
    "["
    u"\U00002700-\U000027BF"  # Dingbats
    u"\U0001F600-\U0001F64F"  # Emoticons
    u"\U00002600-\U000026FF"  # Miscellaneous Symbols
    u"\U0001F300-\U0001F5FF"  # Miscellaneous Symbols And Pictographs
    u"\U0001F900-\U0001F9FF"  # Supplemental Symbols and Pictographs
    u"\U0001FA70-\U0001FAFF"  # Symbols and Pictographs Extended-A
    u"\U0001F680-\U0001F6FF"  # Transport and Map Symbols
    "]+", flags=re.UNICODE)
AUTHOR_REMOVE_HTML = re.compile(r'<[^>]+>')
CLEAN_META_TAGS = re.compile(r'["\']')

PARSER_TAG = re.compile(r"\[[^\]]{1,49}\]")
END_SYMBOL = {'!', '"', "'", ',', '.', ';', '?', '״', '،', '؛', '۔', '।', '॥', '၊', '။', '។', '៕', '’', '‚', '”', '„', '。', '！', '，', '；', '？', '．'}
BLACKLIST_END_SYMBOL = {"…","...","。。。","．．．"}

IMG_SRC_ATTR = [
    "data-src",
    "data-src-fg",
    "data-scroll-image",
    "srcset",
    "data-lateloadsrc",
    "data-img-src",
    "data-original",
    "data-gt-lazy-src",
    "data-lazy",
    "data-lazy-src",
    "src2",
    "src",
]

@lru_cache(maxsize=2**14)  # sys.maxunicode = 1114111
def return_printables_and_spaces(char):
    'Return a character if it belongs to certain classes'
    return char if char.isprintable() or char.isspace() else ''

def remove_control_characters(string):
    '''Prevent non-printable and XML invalid character errors'''
    return ''.join(map(return_printables_and_spaces, string))

def line_processing(line, preserve_space=False, trailing_space=False):
    '''Remove HTML space entities, then discard incompatible unicode
       and invalid XML characters on line level'''
    # spacing HTML entities: https://www.w3.org/MarkUp/html-spec/html-spec_13.html
    # unique code spaces
    new_line = remove_control_characters(line.replace('&#13;', '\r').replace('&#10;', '\n').replace('&nbsp;', '\u00A0'))
    if not preserve_space:
        # remove newlines that are not related to punctuation or markup
        # remove non-printable chars and normalize space characters (including Unicode spaces)
        new_line = trim(LINES_TRIMMING.sub(r" ", new_line))
        # prune empty lines
        if all(map(str.isspace, new_line)):
            new_line = None
        elif trailing_space:
            space_before = " " if line[0] == " " else ""
            space_after = " " if line[-1] == " " else ""
            new_line = "".join([space_before, new_line, space_after])
    return new_line

def normalize_authors(current_authors, author_string):
    '''Normalize author info to focus on author names only'''
    new_authors = []
    if author_string.lower().startswith('http') or AUTHOR_EMAIL.match(author_string):
        return current_authors
    if current_authors is not None:
        new_authors = current_authors.split('; ')
    # fix to code with unicode
    if '\\u' in author_string:
        author_string = author_string.encode().decode('unicode_escape')
    # fix html entities
    if '&#' in author_string or '&amp;' in author_string:
        author_string = unescape(author_string)
    # remove html tags
    author_string = AUTHOR_REMOVE_HTML.sub('', author_string)
    # examine names
    for author in AUTHOR_SPLIT.split(author_string):
        author = trim(author)
        # remove emoji
        author = AUTHOR_EMOJI_REMOVE.sub('', author)
        # remove @username
        author = AUTHOR_TWITTER.sub('', author)
        # replace special characters with space
        author = trim(AUTHOR_REPLACE_JOIN.sub(' ', author))
        author = AUTHOR_REMOVE_NICKNAME.sub('', author)
        # remove special characters
        author = AUTHOR_REMOVE_SPECIAL.sub('', author)
        author = AUTHOR_PREFIX.sub('', author)
        author = AUTHOR_REMOVE_NUMBERS.sub('', author)
        author = AUTHOR_REMOVE_PREPOSITION.sub('', author)
        # skip empty or improbably long strings
        if len(author) == 0 or (
            # simple heuristics, regex or vowel tests also possible
            ' ' not in author and '-' not in author and len(author) >= 50
            ):
            continue
        # title case
        if not author[0].isupper() or sum(1 for c in author if c.isupper()) < 1:
            author = author.title()
        # safety checks
        if author not in new_authors and (len(new_authors) == 0 or all(new_author not in author for new_author in new_authors)):
            new_authors.append(author)
    if len(new_authors) == 0:
        return current_authors
    return '; '.join(new_authors).strip('; ')

def normalize_tags(tags):
    '''Remove special characters of tags'''
    tags = CLEAN_META_TAGS.sub(r'', trim(unescape(tags)))
    return ", ".join(filter(None, tags.split(", ")))

def uniquify_list(l):
    """
    Remove duplicates from a list while keeping order in an efficient way.
    Dictionaries preserve insertion order since Python 3.6.

    https://www.peterbe.com/plog/fastest-way-to-uniquify-a-list-in-python-3.6
    """
    return list(dict.fromkeys(l))

def has_style(style, styles):
    """Does the style string contain any of the styles?
    This function is robust to variations in the spaces between the styles.
    """
    # Remove any spaces.
    style = style.replace(" ", "")
    styles = [s.replace(" ", "") for s in styles]
    for s in styles:
        if s in style:
            return True
    return False


def word_wrap(text, char_width=20):
    """Wrap text to a given width, not breaking words."""
    if not text:
        return ""

    words = text.split()
    lines = []
    current_line = []

    for word in words:
        if len(" ".join(current_line + [word])) <= char_width:
            current_line.append(word)
        else:
            if current_line:  # Check if current_line is not empty
                lines.append(" ".join(current_line))
            current_line = [word]

            # Handle the case when the word is longer than the character width
            while len(current_line[0]) > char_width:
                lines.append(current_line[0][:char_width])
                current_line[0] = current_line[0][char_width:]

    if current_line:
        lines.append(" ".join(current_line))

    return "\n".join(lines)


class ReplacementManager:
    """This replacement manager simply adds tags next to the instances of the text.
    It contains a method to remove these tags."""

    def __init__(self):
        self.tags = set()

    def add_replacement(self, text, tag="default"):
        self.tags.add(tag)
        return f"§§{tag}§§" + text

    def remove_tags(self, text):
        tag_regex = "|".join(f"§§{tag}§§" for tag in self.tags)
        return re.sub(tag_regex, "", text)

    def has_tag(self, text, tag):
        return f"§§{tag}§§" in text


class Config:
    """A simple config object that loads a config from a YAML file and
    presents as a dictionary"""

    def __init__(self, config_file):
        with open(config_file, "r") as f:
            self.config = yaml.safe_load(f)

    def sample_from_list(self, list):
        """Sample from a list of (probability, value) tuples."""
        probabilities = [p for p, _ in list]
        values = [v for _, v in list]
        probabilities = np.array(probabilities)
        probabilities /= probabilities.sum()
        return np.random.choice(values, p=probabilities)

    def _sample(self, config):
        # For every value that has a type of list, first check it is in the format of:
        # - (probability, value)
        # - (probability, value)
        # - ...
        # And the probabilities sum to 1.
        # Then sample from the list.
        sampled_config = {}
        for key, value in config.items():
            # print the type of the value
            if isinstance(value, list):
                # Check the format of the list.
                # Check the probabilities sum to 1.
                # Sample from the list.
                sampled_config[key] = self.sample_from_list(value)
            elif isinstance(value, dict):
                sampled_config[key] = self._sample(value)
            else:
                sampled_config[key] = value
        return sampled_config

    def sample(self):
        return self._sample(self.config)


def trim(string):
    """Remove unnecessary spaces within a text string"""
    try:
        # remove newlines that are not related to punctuation or markup + proper trimming
        # return LINES_TRIMMING.sub(r' ', string).strip(' \t\n\r\v')
        # faster:
        return " ".join(string.split()).strip()
    except (AttributeError, TypeError):
        return None

html2text_parser = html2text.HTML2Text()
html2text_parser.ignore_links = True
html2text_parser.ignore_images = True


def fast_html2text(html, filter_length=None, filter_end_symbol=False, turndown=False):
    text = html2text_parser.handle(html)
    if turndown:
        text = PARSER_TAG.sub("", text)
    if filter_length is not None:
        text = text.split("\n")
        text = [i for i in text if len(i) > filter_length]
        text = "\n".join(text)
    if filter_end_symbol:
        filtered_lines = []

        for line in text.split("\n"):
            if any(line.endswith(blacklist) for blacklist in BLACKLIST_END_SYMBOL):
                continue
            if any(line.endswith(symbol) for symbol in END_SYMBOL):
                filtered_lines.append(line)

        text = "\n".join(filtered_lines)

    return text


from haruka_parser.dictionary import highlightjs

CODING_LANGUAGES = {lang.alias: lang.name for lang in highlightjs.languages}

def extract_code_languages(class_text):
    res = []
    class_list = re.sub(r"[_\- ]", " ", class_text).split()
    for class_name in class_list:
        if class_name in CODING_LANGUAGES.keys():
            res.append(class_name)
    return sorted(res, key=len, reverse=True)



"""Handle the <li>, <ol>, <ul> tags."""
from typing import Dict

from haruka_parser.table_processing.html_state import HtmlDocumentState

UL_COUNTER = ("* ", "+ ", "o ", "- ")
UL_COUNTER_LEN = len(UL_COUNTER)


def get_bullet(state: HtmlDocumentState) -> str:
    """Return the bullet that correspond to the given index."""
    return UL_COUNTER[len(state.li_counter) % UL_COUNTER_LEN]


def li_start_handler(state: HtmlDocumentState, _: Dict) -> None:
    """Handle the <li> tag."""
    bullet = state.li_counter[-1] if state.li_counter else "* "
    if isinstance(bullet, int):
        state.li_counter[-1] += 1
        state.tags[-1].list_bullet = f"{bullet}. "
    else:
        state.tags[-1].list_bullet = bullet

    state.tags[-1].write("")


def ul_start_handler(state: HtmlDocumentState, _: Dict) -> None:
    """Handle the <ul> tag."""
    state.li_counter.append(get_bullet(state))


def ul_end_handler(state: HtmlDocumentState) -> None:
    """Handle the </ul> tag."""
    state.li_counter.pop()


def ol_start_handler(state: HtmlDocumentState, _: Dict) -> None:
    """Handle the <ol> tag."""
    state.li_counter.append(1)


def ol_end_handler(state: HtmlDocumentState) -> None:
    """Handle the </ol> tag."""
    state.li_counter.pop()



"""
Meta-functions to be applied module-wide.
"""

import gc

from courlan.meta import clear_caches as reset_caches_courlan
from htmldate.meta import reset_caches as reset_caches_htmldate
from justext.core import define_stoplist

from .deduplication import LRU_TEST, Simhash, is_similar_domain
from .utils import line_processing, return_printables_and_spaces, trim


def reset_caches() -> None:
    """Reset all known LRU caches used to speed up processing.
       This may release some memory."""
    # justext
    define_stoplist.cache_clear()
    # handles htmldate and charset_normalizer
    reset_caches_htmldate()
    # courlan
    reset_caches_courlan()
    # own
    is_similar_domain.cache_clear()
    line_processing.cache_clear()
    return_printables_and_spaces.cache_clear()
    trim.cache_clear()
    LRU_TEST.clear()
    Simhash._vector_to_add.cache_clear()
    # garbage collection
    gc.collect()



import os
from setuptools import setup

def package_files(directory):
    paths = []
    for root, dirs, files in os.walk(directory):
        for file in files:
            cur_file = os.path.join('..', root, file)
            if any(exclude in cur_file for exclude in ['/tests/', '/docs/', '/.git', '__pycache__', '.html']):
                continue
            paths.append(cur_file)
    return paths

# 0.1 version from OpenWebMath: https://github.com/keirp/OpenWebMath
# OpenWebMath is made available under an ODC-By 1.0 license; users should also abide by the CommonCrawl ToU: https://commoncrawl.org/terms-of-use/. We do not alter the license of any of the underlying data.
setup(
    name="haruka_parser",
    version="0.6.4",
    description="A simple HTML Parser",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="papersnake",
    author_email="prnake@gmail.com",
    url="https://github.com/prnake/haruka-parser",
    packages=["haruka_parser"],
    package_data={'': package_files("haruka_parser")},
    install_requires=[
        "py_asciimath",
        "inscriptis",
        "tabulate",
        "numpy",
        "resiliparse",
        "ftfy",
        "faust-cchardet",
        "lxml",
        "lxml_html_clean",
        "courlan",
        "html2text",
        "charset_normalizer",
        "trafilatura",
    ],
    include_package_data=True,
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Text Processing",
        "Topic :: Text Processing :: Markup :: HTML",
        "Topic :: Utilities",
    ),
)



from collections import defaultdict
from copy import deepcopy

from lxml.etree import XPath, strip_tags

from .utils import IMG_SRC_ATTR, trim, uniquify_list

# order could matter, using lists to keep extraction deterministic
MANUALLY_CLEANED = [
    # important
    'aside', 'embed', 'footer', 'form', 'head', 'iframe', 'menu',
    # other content
    'applet', 'canvas', 'map',
    # secondary
    'area', 'blink', 'button', 'datalist', 'dialog',
    'frame', 'frameset', 'fieldset', 'link', 'input', 'ins', 'label', 'legend',
    'marquee', 'menuitem', 'nav', 'optgroup', 'option',
    'output', 'param', 'progress', 'rp', 'rt', 'rtc', 'select',
    'style', 'track', 'textarea', 'time', 'use',
]
# 'meta', 'hr', 'img', 'data', 'details', 'summary', 'math', 'script'

MANUALLY_STRIPPED = [
    'abbr', 'acronym', 'address', 'bdi', 'bdo', 'big', 'cite', 'data', 'dfn',
    'font', 'hgroup', 'ins', 'kbd', 'mark', 'meta', 'ruby', 'small', 'tbody',
    'template', 'tfoot', 'thead', 's', 'samp', 'shadow', 'small', 'strike',
    'tt', 'u', 'var'
]
# 'center', 'rb', 'wbr'

# filters
CUT_EMPTY_ELEMS = {'article', 'b', 'blockquote', 'dd', 'div', 'dt', 'em',
                   'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'i', 'li', 'main',
                   'p', 'pre', 'q', 'section', 'span', 'strong'}
                   # 'meta', 'td', 'a', 'caption', 'dl', 'header',
                   # 'colgroup', 'col',
#CUT_EMPTY_ELEMS = {'div', 'span'}

# learn from obelics
MEDIA_CONTAIN_INTERESTING_ATTRIBUTES_SET = [
    "audio", "embed", "figure", "iframe",
    "img", "object", "picture", "video", "source",
]

OVERALL_DISCARD_XPATH = [XPath(x) for x in (
    # navigation + footers, news outlets related posts, sharing, jp-post-flair jp-relatedposts
    '''.//*[(self::div or self::item or self::list
             or self::p or self::section or self::span)][
    contains(translate(@id, "F","f"), "footer") or contains(translate(@class, "F","f"), "footer")
    or contains(@id, "related") or contains(translate(@class, "R", "r"), "related") or
    contains(@id, "viral") or contains(@class, "viral") or
    starts-with(@id, "shar") or starts-with(@class, "shar") or
    contains(@class, "share-") or
    contains(translate(@id, "S", "s"), "share") or
    contains(@id, "social") or contains(@class, "social") or contains(@class, "sociable") or
    contains(@id, "syndication") or contains(@class, "syndication") or
    starts-with(@id, "jp-") or starts-with(@id, "dpsp-content") or
    contains(@class, "embedded") or contains(@class, "embed")
    or contains(@id, "newsletter") or contains(@class, "newsletter")
    or contains(@class, "subnav") or
    contains(@id, "cookie") or contains(@class, "cookie") or contains(@id, "tags")
    or contains(@class, "tags")  or contains(@id, "sidebar") or
    contains(@class, "sidebar") or contains(@id, "banner") or contains(@class, "banner")
    or contains(@class, "meta") or
    contains(@id, "menu") or contains(@class, "menu") or
    contains(translate(@id, "N", "n"), "nav") or contains(translate(@role, "N", "n"), "nav")
    or starts-with(@class, "nav") or contains(translate(@class, "N", "n"), "navigation") or
    contains(@class, "navbar") or contains(@class, "navbox") or starts-with(@class, "post-nav")
    or contains(@id, "breadcrumb") or contains(@class, "breadcrumb") or
    contains(@id, "bread-crumb") or contains(@class, "bread-crumb") or
    contains(@id, "author") or contains(@class, "author") or
    contains(@id, "button") or contains(@class, "button")
    or contains(translate(@class, "B", "b"), "byline")
    or contains(@class, "rating") or starts-with(@class, "widget") or
    contains(@class, "attachment") or contains(@class, "timestamp") or
    contains(@class, "user-info") or contains(@class, "user-profile") or
    contains(@class, "-ad-") or contains(@class, "-icon")
    or contains(@class, "article-infos") or
    contains(translate(@class, "I", "i"), "infoline")
    or contains(@data-component, "MostPopularStories")
    or contains(@class, "outbrain") or contains(@class, "taboola")
    or contains(@class, "criteo") or contains(@class, "options")
    or contains(@class, "consent") or contains(@class, "modal-content")
    or contains(@class, "paid-content") or contains(@class, "paidcontent")
    or contains(@id, "premium-") or contains(@id, "paywall")
    or contains(@class, "obfuscated") or contains(@class, "blurred")
    or contains(@class, " ad ")
    or contains(@class, "next-post") or contains(@class, "side-stories")
    or contains(@class, "related-stories") or contains(@class, "most-popular")
    or contains(@class, "mol-factbox") or starts-with(@class, "ZendeskForm")
    or contains(@class, "message-container") or contains(@id, "message_container")
    or contains(@class, "yin") or contains(@class, "zlylin") or
    contains(@class, "xg1") or contains(@id, "bmdh")
    or @data-lp-replacement-content or @data-testid]''',

    # comment debris + hidden parts
    '''.//*[@class="comments-title" or contains(@class, "comments-title") or
    contains(@class, "nocomments") or starts-with(@id, "reply-") or starts-with(@class, "reply-") or
    contains(@class, "-reply-") or contains(@class, "message") or contains(@id, "reader-comments")
    or contains(@id, "akismet") or contains(@class, "akismet") or contains(@class, "suggest-links") or
    starts-with(@class, "hide-") or contains(@class, "hide-print") or contains(@id, "hidden")
    or contains(@style, "hidden") or contains(@hidden, "hidden") or contains(@class, "noprint")
    or contains(@style, "display:none") or contains(@style, "display: none") or contains(@class, " hidden") or @aria-hidden="true"
    or contains(@class, "notloaded")]''',
)]
# conflicts:
# contains(@id, "header") or contains(@class, "header") or
# class contains "cats" (categories, also tags?)
# or contains(@class, "hidden ")  or contains(@class, "-hide")
# or contains(@class, "paywall")
# contains(@class, "content-info") or contains(@class, "content-title")
# contains(translate(@class, "N", "n"), "nav") or

PAYWALL_DISCARD_XPATH = [XPath(
    '''.//*[(self::div or self::p)][
    contains(@id, "paywall") or contains(@id, "premium") or
    contains(@class, "paid-content") or contains(@class, "paidcontent") or
    contains(@class, "obfuscated") or contains(@class, "blurred") or
    contains(@class, "restricted") or contains(@class, "overlay")
    ]'''
)]

# the following conditions focus on extraction precision
TEASER_DISCARD_XPATH = [XPath(
    '''.//*[(self::div or self::item or self::list
             or self::p or self::section or self::span)][
        contains(translate(@id, "T", "t"), "teaser") or contains(translate(@class, "T", "t"), "teaser")
    ]'''
)]

def delete_element(element):
    "Remove the element from the LXML tree."
    try:
        element.drop_tree()  # faster when applicable
    except AttributeError:  # pragma: no cover
        element.getparent().remove(element)


def tree_cleaning(tree):
    "Prune the tree by discarding unwanted elements."
    # determine cleaning strategy, use lists to keep it deterministic
    cleaning_list, stripping_list = \
        MANUALLY_CLEANED.copy(), MANUALLY_STRIPPED.copy()
    
    # prevent this issue: https://github.com/adbar/trafilatura/issues/301
    for elem in tree.xpath('.//figure[descendant::table]'):
        elem.tag = 'div'
        
    # strip targeted elements
    strip_tags(tree, stripping_list)

    # delete targeted elements
    for expression in cleaning_list:
        for element in tree.getiterator(expression):
            delete_element(element)
            
    tree = remove_empty_nodes(tree)

    return prune_html(tree)


def prune_html(tree):
    "Delete selected empty elements to save space and processing time."
    # //comment() needed for date extraction
    for element in tree.xpath("//processing-instruction()|//*[not(node())]"):
        if element.tag in CUT_EMPTY_ELEMS:
            delete_element(element)
    return tree

def prune_unwanted_nodes(tree, nodelist, with_backup=False):
    '''Prune the HTML tree by removing unwanted sections.'''
    if with_backup is True:
        old_len = len(tree.text_content())  # ' '.join(tree.itertext())
        backup = deepcopy(tree)
    for expression in nodelist:
        for subtree in expression(tree):
            # preserve tail text from deletion
            if subtree.tail is not None:
                previous = subtree.getprevious()
                if previous is None:
                    previous = subtree.getparent()
                if previous is not None:
                    # There is a previous node, append text to its tail
                    if previous.tail is not None:
                        previous.tail = ' '.join([previous.tail, subtree.tail])
                    else:
                        previous.tail = subtree.tail
            # remove the node
            subtree.getparent().remove(subtree)
    if with_backup is False:
        return tree
    # else:
    new_len = len(tree.text_content())
    # todo: adjust for recall and precision settings
    if new_len > old_len/7:
        return tree
    return backup

def collect_link_info(links_xpath, favor_precision=False):
    '''Collect heuristics on link text'''
    # init
    shortelems, mylist = 0, []
    # longer strings impact recall in favor of precision
    threshold = 10 if not favor_precision else 50
    # examine the elements
    for subelem in links_xpath:
        subelemtext = trim(subelem.text_content())
        if subelemtext:
            mylist.append(subelemtext)
            if len(subelemtext) < threshold:
                shortelems += 1
    lengths = sum(len(text) for text in mylist)
    return lengths, len(mylist), shortelems, mylist

def link_density_test(element, text, favor_precision=False):
    '''Remove sections which are rich in links (probably boilerplate)'''
    links_xpath, mylist = element.findall('.//ref'), []
    if links_xpath:
        if element.tag == 'p': #  and not element.getparent().tag == 'item'
            if favor_precision is False:
                if element.getnext() is None:
                    limitlen, threshold = 60, 0.8
                else:
                    limitlen, threshold = 30, 0.8
            else:
                limitlen, threshold = 200, 0.8
            #if 'hi' in list(element):
            #    limitlen, threshold = 100, 0.8
        #elif element.tag == 'head':
        #    limitlen, threshold = 50, 0.8
        else:
            if element.getnext() is None:
                limitlen, threshold = 300, 0.8
            #elif re.search(r'[.?!:]', elemtext):
            #    limitlen, threshold = 150, 0.66
            else:
                limitlen, threshold = 100, 0.8
        elemlen = len(text)
        if elemlen < limitlen:
            linklen, elemnum, shortelems, mylist = collect_link_info(links_xpath, favor_precision)
            if elemnum == 0:
                return True, mylist
            # (elemnum > 1 and shortelems/elemnum > 0.8):
            if linklen > threshold*elemlen or (elemnum > 1 and shortelems/elemnum > 0.8):
                return True, mylist
    return False, mylist

def delete_by_link_density(subtree, tagname, backtracking=False, favor_precision=False):
    '''Determine the link density of elements with respect to their length,
       and remove the elements identified as boilerplate.'''
    myelems, deletions = defaultdict(list), []
    for elem in subtree.iter(tagname):
        elemtext = trim(elem.text_content())
        result, templist = link_density_test(elem, elemtext, favor_precision)
        if result is True:
            deletions.append(elem)
        elif backtracking is True and len(templist) > 0:  # if?
            myelems[elemtext].append(elem)
    # summing up
    if backtracking is True:
        if favor_precision is False:
            threshold = 100
        else:
            threshold = 200
        for text, elem in myelems.items():
            if 0 < len(text) < threshold and len(elem) >= 3:
                deletions.extend(elem)
                # print('backtrack:', text)
            # else: # and not re.search(r'[?!.]', text):
            # print(elem.tag, templist)
    for elem in uniquify_list(deletions):
        try:
            elem.getparent().remove(elem)
        except AttributeError:
            pass
    return subtree

def get_media_src(node):
    node_attributes = node.attrib
    node_tag = node.tag
    src = None

    if node_tag == "img":
        # Check all possible source type, and keep the first valid one
        for source_type in IMG_SRC_ATTR:
            if source_type in node_attributes and node_attributes[source_type]:
                if ("," not in node_attributes[source_type]) and (" " not in node_attributes[source_type]):
                    src = node_attributes[source_type]
                    break

    elif node_tag == "video":
        if ("src" in node_attributes) and node_attributes["src"]:
            src = node_attributes["src"]
        else:
            for cnode in node.iterchildren():
                if not src:
                    if cnode.tag == "source":
                        cnode_attributes = cnode.attrib
                        if ("src" in cnode_attributes) and cnode_attributes["src"]:
                            src = cnode_attributes["src"]

    elif node_tag == "audio":
        if ("src" in node_attributes) and node_attributes["src"]:
            src = node_attributes["src"]
        else:
            for cnode in node.iterchildren():
                if not src:
                    if cnode.tag == "source":
                        cnode_attributes = cnode.attrib
                        if ("src" in cnode_attributes) and cnode_attributes["src"]:
                            src = cnode_attributes["src"]
    else:
        return None  # TODO iframes

    # Check on comma because it's non-canonical and should not be used anyway in urls.
    # TODO: have checks on valid URLs
    # Useless (at least for images) since already checked
    if src is not None and (("," in src) or (" " in src)):
        return None

    return src


def remove_empty_nodes(tree):
    """
    Function used to remove empty leaves iteratively, so it also ends up also removing nodes
    that are higher up in the tree.
    """
    modification = True
    while modification:
        nodes_to_remove = [
            node
            for node in tree.getroot().iter()
            if (
                (node.tag not in MEDIA_CONTAIN_INTERESTING_ATTRIBUTES_SET)
                and (not [child for child in node.iterchildren()])
                and (not node.text.strip())
                and (node.tag != "html")
            )
            or (
                (node.tag in MEDIA_CONTAIN_INTERESTING_ATTRIBUTES_SET)
                and not get_media_src(node)
            )
        ]
        if nodes_to_remove:
            for node in nodes_to_remove:
                node.getparent().remove(node)
        else:
            modification = False
    return tree